import inspect
from typing import Any, Optional

import pydantic

from pjrpc.common.typedefs import JsonRpcParamsT
from pjrpc.server.dispatcher import Method

from . import base
from .base import ExcludeFunc, MethodType


class PydanticValidatorFactory(base.BaseValidatorFactory):
    """
    Method parameters validator factory based on `pydantic <https://pydantic-docs.helpmanual.io/>`_ library.
    Uses python type annotations for parameters validation.

    :param exclude: a function that decides if the parameters must be excluded
                    from validation (useful for dependency injection)
    """

    def __init__(self, exclude: Optional[ExcludeFunc] = None, **config_args: Any):
        super().__init__(exclude=exclude)

        config_args.setdefault('extra', 'forbid')
        self._model_config = config_args

    def __call__(self, method: Method) -> Method:
        self.build(method.func)
        return method

    def build(self, method: MethodType) -> 'PydanticValidator':
        return PydanticValidator(method, self._model_config, self._exclude)


class PydanticValidator(base.BaseValidator):
    """
    Pydantic method parameters validator based on `pydantic <https://docs.pydantic.dev/latest/>`_ library.
    """

    def __init__(
        self,
        method: MethodType,
        model_config: dict[str, Any],
        exclude: Optional[ExcludeFunc] = None,
    ):
        super().__init__(method, exclude)
        self._model_config = model_config
        self._params_model = self._build_validation_model(method.__name__)

    def validate_params(self, params: Optional['JsonRpcParamsT']) -> dict[str, Any]:
        """
        Validates params against method using ``pydantic`` validator.

        :param params: parameters to be validated
        """

        model_fields: tuple[str, ...] = tuple(self._params_model.__fields__)  # type: ignore[arg-type]

        if isinstance(params, dict):
            params_dict = params
        elif isinstance(params, (list, tuple)):
            if len(params) > len(fields := list(model_fields)):
                fields.extend((f'params.{n}' for n in range(len(fields), len(params) + 1)))
            params_dict = {name: value for name, value in zip(fields, params)}
        else:
            raise AssertionError("unreachable")

        try:
            obj = self._params_model.parse_obj(params_dict)
        except pydantic.ValidationError as e:
            raise base.ValidationError(*e.errors()) from e

        return {field_name: obj.__dict__[field_name] for field_name in fields}

    def _build_validation_model(self, method_name: str) -> type[pydantic.BaseModel]:
        schema = self._build_validation_schema(self._signature)
        return pydantic.create_model(method_name, **schema, __config__=pydantic.config.get_config(self._model_config))

    def _build_validation_schema(self, signature: inspect.Signature) -> dict[str, Any]:
        """
        Builds pydantic model based validation schema from method signature.

        :param signature: method signature to build schema for
        :returns: validation schema
        """

        field_definitions: dict[str, tuple[Any, Any]] = {}

        for param in signature.parameters.values():
            if param.kind is inspect.Parameter.VAR_KEYWORD:
                field_definitions[param.name] = (
                    Optional[dict[str, param.annotation]]  # type: ignore
                    if param.annotation is not inspect.Parameter.empty else Any,
                    param.default if param.default is not inspect.Parameter.empty else None,
                )
            elif param.kind is inspect.Parameter.VAR_POSITIONAL:
                field_definitions[param.name] = (
                    Optional[dict[param.annotation]]  # type: ignore
                    if param.annotation is not inspect.Parameter.empty else Any,
                    param.default if param.default is not inspect.Parameter.empty else None,
                )
            else:
                field_definitions[param.name] = (
                    param.annotation if param.annotation is not inspect.Parameter.empty else Any,
                    param.default if param.default is not inspect.Parameter.empty else ...,
                )

        return field_definitions



import abc
import contextlib as cl
import functools as ft
import json
import logging
from typing import Any, AsyncGenerator, Awaitable, Callable, Generator, Iterable, Mapping, Optional, Protocol, TypeVar

from pjrpc import common
from pjrpc.common import UNSET, AbstractRequest, AbstractResponse, BatchRequest, BatchResponse, MaybeSet, Request
from pjrpc.common import Response, exceptions, generators
from pjrpc.common.typedefs import JsonRpcRequestIdT, JsonT

logger = logging.getLogger(__package__)


ReturnT = TypeVar('ReturnT', covariant=True)


class ProxyCall(Protocol[ReturnT]):
    def __call__(self, *args: JsonT, **kwargs: JsonT) -> ReturnT: pass


class Batch:
    """
    Batch object. Provides syntactic sugar to send batch requests.
    """

    class Proxy:
        """
        Proxy object. Provides syntactic sugar to make method call using dot notation.

        :param batch: batch object
        """

        def __init__(self, batch: 'Batch'):
            self._batch = batch

        def __getattr__(self, attr: str) -> ProxyCall[JsonT]:
            return ft.partial(self._batch.call, attr)

    def __init__(
        self,
        id_gen_impl: Callable[..., Generator[JsonRpcRequestIdT, None, None]],
    ):
        self._id_gen = id_gen_impl()

        self._requests: list[Request] = []
        self._response: MaybeSet[Optional[BatchResponse]] = UNSET

    @property
    def proxy(self) -> Proxy:
        """
        Client proxy object.
        """

        return Batch.Proxy(self)

    @property
    def requests(self) -> list[Request]:
        """
        Batch requests.
        """

        return self._requests

    def __call__(self, method: str, *args: JsonT, **kwargs: JsonT) -> None:
        """
        Makes a JSON-RPC call.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :returns: response result
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        self._requests.append(Request(id=next(self._id_gen), method=method, params=args or kwargs))

    def send(self, request: Request) -> None:
        self._requests.append(request)

    def notify(self, method: str, *args: JsonT, **kwargs: JsonT) -> None:
        """
        Makes a notification request

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        self._requests.append(Request(id=None, method=method, params=args or kwargs))

    def call(self, method: str, *args: JsonT, **kwargs: JsonT) -> None:
        """
        Makes a JSON-RPC call.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :returns: response result
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        self._requests.append(Request(id=next(self._id_gen), method=method, params=args or kwargs))

    def set_response(self, response: Optional[BatchResponse]) -> None:
        """
        Sets batch response
        """

        self._response = response

    def get_response(self) -> Optional[BatchResponse]:
        """
        Returns a batch response.
        """

        if self._response is UNSET:
            raise RuntimeError("batch reqeust is not sent yet")

        return self._response

    def get_results(self) -> Iterable[Any]:
        """
        Returns the batch results preserving requests order (skipping notification request).
        """

        if self._response is UNSET:
            raise RuntimeError("batch reqeust is not sent yet")
        if self._response is None:
            return []

        if self._response.is_error:
            raise self._response.unwrap_error()

        response_map = {response.id: response for response in self._response}
        results: list[Any] = []
        for request in self._requests:
            if request.id is not None:
                if (response := response_map.get(request.id)) is None:
                    raise exceptions.IdentityError(f"response '{request.id}' is missing")
                results.append(response.unwrap_result())

        return results


MiddlewareHandler = Callable[[AbstractRequest, Mapping[str, Any]], Optional[AbstractResponse]]


class Middleware(Protocol):
    """
    JSON-RPC client middleware.
    """

    def __call__(
        self,
        request: AbstractRequest,
        request_kwargs: Mapping[str, Any],
        /,
        handler: MiddlewareHandler,
    ) -> Optional[AbstractResponse]:
        pass


class AbstractClient(abc.ABC):
    """
    Abstract synchronous JSON-RPC client.

    :param id_gen_impl: identifier generator
    :param error_cls: JSON-RPC error base class
    :param json_loader: json loader
    :param json_dumper: json dumper
    :param json_encoder: json encoder
    :param json_decoder: json decoder
    :param middlewares: client reqeust middlewares
    """

    class Proxy:
        """
        Proxy object. Provides syntactic sugar to make method call using dot notation.

        :param client: JSON-RPC client instance
        """

        def __init__(self, client: 'AbstractClient'):
            self._client = client

        def __getattr__(self, attr: str) -> ProxyCall[JsonT]:
            return ft.partial(self._client.call, attr)

    @property
    def proxy(self) -> Proxy:
        """
        Client proxy object.
        """

        return AbstractClient.Proxy(self)

    @cl.contextmanager
    def batch(self) -> Generator[Batch, None, None]:
        """
        Client batch wrapper.
        """

        batch = Batch(self._id_gen_impl)
        yield batch

        response = self._send(BatchRequest(*batch.requests), {})
        assert isinstance(response, (BatchResponse, type(None))), "unexpected response type"

        batch.set_response(response)

    def __init__(
        self,
        *,
        id_gen_impl: Callable[..., Generator[JsonRpcRequestIdT, None, None]] = generators.sequential,
        error_cls: type[exceptions.JsonRpcError] = exceptions.JsonRpcError,
        json_loader: Callable[..., Any] = json.loads,
        json_dumper: Callable[..., str] = json.dumps,
        json_encoder: type[common.JSONEncoder] = common.JSONEncoder,
        json_decoder: Optional[json.JSONDecoder] = None,
        middlewares: Iterable[Middleware] = (),
        request_content_type: str = common.DEFAULT_CONTENT_TYPE,
        response_content_types: Iterable[str] = common.RESPONSE_CONTENT_TYPES,
    ):
        self._id_gen_impl = id_gen_impl
        self._error_cls = error_cls
        self._json_loader = json_loader
        self._json_dumper = json_dumper
        self._json_encoder = json_encoder
        self._json_decoder = json_decoder
        self._request_content_type = request_content_type
        self._response_content_types = set(response_content_types)

        send = self._send_request
        for middleware in reversed(list(middlewares)):
            send = ft.partial(middleware, handler=send)

        self._send = send

    def __call__(self, method: str, *args: JsonT, **kwargs: JsonT) -> JsonT:
        """
        Makes a JSON-RPC call.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :returns: response result
        """

        return self.call(method, *args, **kwargs)

    @abc.abstractmethod
    def _request(
        self,
        request_text: str,
        is_notification: bool,
        request_kwargs: Mapping[str, Any],
    ) -> Optional[str]:
        """
        Makes a JSON-RPC request.

        :param request_text: request text representation
        :param is_notification: is the request a notification
        :param request_kwargs: additional client request argument
        :returns: response text representation
        """

    def _send_request(
        self,
        request: AbstractRequest,
        request_kwargs: Mapping[str, Any],
    ) -> Optional[AbstractResponse]:
        """
        Sends a JSON-RPC request.

        :param request: request instance
        :param request_kwargs: additional client request argument
        :returns: response instance or None if the request is a notification
        """

        response_cls: type[AbstractResponse] = BatchResponse if isinstance(request, BatchRequest) else Response
        request_text = self._json_dumper(request, cls=self._json_encoder)

        response_text = self._request(request_text, request.is_notification, request_kwargs)
        if not request.is_notification:
            response = response_cls.from_json(
                self._json_loader(response_text, cls=self._json_decoder), error_cls=self._error_cls,
            )
        else:
            if response_text:
                raise exceptions.ProtocolError("unexpected response")
            response = None

        return response

    def notify(self, method: str, *args: JsonT, **kwargs: JsonT) -> None:
        """
        Makes a notification request

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        request = Request(
            id=None,
            method=method,
            params=args or kwargs,
        )
        self._send(request, {})

    def call(self, method: str, *args: JsonT, **kwargs: JsonT) -> JsonT:
        """
        Makes a JSON-RPC call.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :returns: response result
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        request = Request(
            id=next(self._id_gen_impl()),
            method=method,
            params=args or kwargs,
        )
        response = self._send(request, {})

        assert response is not None, "response is not set"
        return response.unwrap_result()


AsyncMiddlewareHandler = Callable[[AbstractRequest, Mapping[str, Any]], Awaitable[Optional[AbstractResponse]]]


class AsyncMiddleware(Protocol):
    """
    Asynchronous JSON-RPC client middleware.
    """

    async def __call__(
        self,
        request: AbstractRequest,
        request_kwargs: Mapping[str, Any],
        /,
        handler: AsyncMiddlewareHandler,
    ) -> Optional[AbstractResponse]:
        pass


class AbstractAsyncClient(abc.ABC):
    """
    Abstract asynchronous JSON-RPC client.

    :param id_gen_impl: identifier generator
    :param error_cls: JSON-RPC error base class
    :param json_loader: json loader
    :param json_dumper: json dumper
    :param json_encoder: json encoder
    :param json_decoder: json decoder
    :param middlewares: client reqeust middlewares
    """

    class Proxy:
        """
        Proxy object. Provides syntactic sugar to make method call using dot notation.

        :param client: JSON-RPC client instance
        """

        def __init__(self, client: 'AbstractAsyncClient'):
            self._client = client

        def __getattr__(self, attr: str) -> Callable[..., Awaitable[JsonT]]:
            return ft.partial(self._client.call, attr)

    @property
    def proxy(self) -> Proxy:
        """
        Client proxy object.
        """

        return AbstractAsyncClient.Proxy(self)

    @cl.asynccontextmanager
    async def batch(self) -> AsyncGenerator[Batch, None]:
        """
        Client batch wrapper.
        """

        batch = Batch(self._id_gen_impl)
        yield batch

        response = await self._send(BatchRequest(*batch.requests), {})
        assert isinstance(response, (BatchResponse, type(None))), "unexpected response type"

        batch.set_response(response)

    def __init__(
        self,
        *,
        id_gen_impl: Callable[..., Generator[JsonRpcRequestIdT, None, None]] = generators.sequential,
        error_cls: type[exceptions.JsonRpcError] = exceptions.JsonRpcError,
        json_loader: Callable[..., Any] = json.loads,
        json_dumper: Callable[..., str] = json.dumps,
        json_encoder: type[common.JSONEncoder] = common.JSONEncoder,
        json_decoder: Optional[json.JSONDecoder] = None,
        middlewares: Iterable[AsyncMiddleware] = (),
        request_content_type: str = common.DEFAULT_CONTENT_TYPE,
        response_content_types: Iterable[str] = common.RESPONSE_CONTENT_TYPES,
    ):
        self._id_gen_impl = id_gen_impl
        self._error_cls = error_cls
        self._json_loader = json_loader
        self._json_dumper = json_dumper
        self._json_encoder = json_encoder
        self._json_decoder = json_decoder
        self._request_content_type = request_content_type
        self._response_content_types = set(response_content_types)

        send = self._send_request
        for middleware in reversed(list(middlewares)):
            send = ft.partial(middleware, handler=send)

        self._send = send

    def __call__(self, method: str, *args: JsonT, **kwargs: JsonT) -> Awaitable[JsonT]:
        """
        Makes a JSON-RPC call.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :returns: response result
        """

        return self.call(method, *args, **kwargs)

    @abc.abstractmethod
    async def _request(
        self,
        request_text: str,
        is_notification: bool,
        request_kwargs: Mapping[str, Any],
    ) -> Optional[str]:
        """
        Makes a JSON-RPC request.

        :param request_text: request text representation
        :param is_notification: is the request a notification
        :param request_kwargs: additional client request argument
        :returns: response text representation or None if the request is a notification
        """

    async def _send_request(
        self,
        request: AbstractRequest,
        request_kwargs: Mapping[str, Any],
    ) -> Optional[AbstractResponse]:
        """
        Sends a JSON-RPC request.

        :param request: request instance
        :param request_kwargs: additional client request argument
        :returns: response instance or None if the request is a notification
        """

        response_cls: type[AbstractResponse] = BatchResponse if isinstance(request, BatchRequest) else Response
        request_text = self._json_dumper(request, cls=self._json_encoder)

        response_text = await self._request(request_text, request.is_notification, request_kwargs)
        if not request.is_notification:
            response = response_cls.from_json(
                self._json_loader(response_text, cls=self._json_decoder), error_cls=self._error_cls,
            )
        else:
            if response_text:
                raise exceptions.ProtocolError("unexpected response")
            response = None

        return response

    async def notify(self, method: str, *args: JsonT, **kwargs: JsonT) -> None:
        """
        Makes a notification request

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        request = Request(
            id=None,
            method=method,
            params=args or kwargs,
        )
        await self._send(request, {})

    async def call(self, method: str, *args: JsonT, **kwargs: JsonT) -> JsonT:
        """
        Makes a JSON-RPC call.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :returns: response result
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        request = Request(
            id=next(self._id_gen_impl()),
            method=method,
            params=args or kwargs,
        )
        response = await self._send(request, {})

        assert response is not None, "response is empty"
        return response.unwrap_result()



"""
JSON-RPC server package.
"""

from . import typedefs
from .dispatcher import AsyncDispatcher, BaseDispatcher, Dispatcher, JSONEncoder, Method, MethodRegistry
from .typedefs import AsyncHandlerType, AsyncMiddlewareType, HandlerType, MiddlewareType
from .utils import exclude_named_param, exclude_positional_param

__all__ = [
    'AsyncDispatcher',
    'AsyncHandlerType',
    'AsyncMiddlewareType',
    'BaseDispatcher',
    'Dispatcher',
    'exclude_named_param',
    'exclude_positional_param',
    'HandlerType',
    'JSONEncoder',
    'Method',
    'MethodRegistry',
    'MiddlewareType',
    'typedefs',
]



"""
OpenAPI Specification generator. See https://swagger.io/specification/.
"""

import dataclasses as dc
from collections import defaultdict
from typing import Any, Callable, Iterable, Mapping, Optional, Union

from pjrpc.common import UNSET, MaybeSet, UnsetType, exceptions
from pjrpc.server import Method, utils
from pjrpc.server.specs import Specification
from pjrpc.server.specs.extractors import BaseMethodInfoExtractor
from pjrpc.server.specs.schemas import build_request_schema, build_response_schema

from .spec import ApiKeyLocation, Components, Contact, Encoding, ExampleObject, ExternalDocumentation, Header, Info
from .spec import Json, JsonSchema, License, Link, MediaType, MethodExample, OAuthFlow, OAuthFlows, Operation
from .spec import Parameter, ParameterLocation, Path, Reference, RequestBody, Response, SecurityScheme
from .spec import SecuritySchemeType, Server, ServerVariable, SpecRoot, StyleType, Tag

__all__ = [
    'ApiKeyLocation',
    'Components',
    'Contact',
    'Encoding',
    'ExampleObject',
    'ExternalDocumentation',
    'Header',
    'Info',
    'Json',
    'JsonSchema',
    'License',
    'Link',
    'MediaType',
    'metadata',
    'MethodExample',
    'MethodMetadata',
    'OAuthFlow',
    'OAuthFlows',
    'OpenAPI',
    'Operation',
    'Parameter',
    'ParameterLocation',
    'Path',
    'Reference',
    'RequestBody',
    'Response',
    'SecurityScheme',
    'SecuritySchemeType',
    'Server',
    'ServerVariable',
    'SpecRoot',
    'StyleType',
    'Tag',
]


HTTP_DEFAULT_STATUS = 200
JSONRPC_MEDIATYPE = 'application/json'


def drop_unset(obj: Any) -> Any:
    if isinstance(obj, dict):
        return dict((drop_unset(k), drop_unset(v)) for k, v in obj.items() if k is not UNSET and v is not UNSET)
    if isinstance(obj, (tuple, list, set)):
        return list(drop_unset(v) for v in obj if v is not UNSET)

    return obj


MethodType = Callable[..., Any]


class OpenAPI(Specification):
    """
    OpenAPI Specification.

    :param info: provides metadata about the API
    :param servers: an array of Server Objects, which provide connectivity information to a target server
    :param external_docs: additional external documentation
    :param openapi: the semantic version number of the OpenAPI Specification version that the OpenAPI document uses
    :param json_schema_dialect: the default value for the $schema keyword within Schema Objects
    :param tags: a list of tags used by the specification with additional metadata
    :param security: a declaration of which security mechanisms can be used across the API
    :param security_schemes: an object to hold reusable Security Scheme Objects
    """

    def __init__(
        self,
        info: Info,
        servers: MaybeSet[list[Server]] = UNSET,
        external_docs: MaybeSet[ExternalDocumentation] = UNSET,
        tags: MaybeSet[list[Tag]] = UNSET,
        security: MaybeSet[list[dict[str, list[str]]]] = UNSET,
        security_schemes: MaybeSet[dict[str, Union[SecurityScheme, Reference]]] = UNSET,
        openapi: str = '3.1.0',
        json_schema_dialect: MaybeSet[str] = UNSET,
    ):
        self._info = info
        self._servers = servers
        self._external_docs = external_docs
        self._tags = tags
        self._security = security
        self._security_schemes = security_schemes
        self._openapi = openapi
        self._json_schema_dialect = json_schema_dialect

    def generate(self, root_endpoint: str, methods: Mapping[str, Iterable[Method]]) -> dict[str, Any]:
        spec_root = SpecRoot(
            info=self._info,
            paths={},
            components=Components(securitySchemes=self._security_schemes),
            servers=self._servers,
            externalDocs=self._external_docs,
            tags=self._tags,
            security=self._security,
            openapi=self._openapi,
            jsonSchemaDialect=self._json_schema_dialect,
        )
        spec_root.components.schemas = {}

        methods_list = [(path, method) for path, methods in methods.items() for method in methods]
        for endpoint, method in methods_list:
            path = utils.join_path(root_endpoint, endpoint)
            for meta in method.metadata:
                if isinstance(meta, MethodSpecification):
                    spec_root.paths[f'{path}#{method.name}'] = Path(post=meta.operation)
                    spec_root.components.schemas.update(meta.component_schemas or {})

        return drop_unset(dc.asdict(spec_root))


@dc.dataclass(frozen=True)
class MethodMetadata:
    params_schema: MaybeSet[dict[str, JsonSchema]] = UNSET
    result_schema: MaybeSet[JsonSchema] = UNSET
    examples: MaybeSet[list[MethodExample]] = UNSET
    tags: MaybeSet[list[Tag]] = UNSET
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    external_docs: MaybeSet[ExternalDocumentation] = UNSET
    deprecated: MaybeSet[bool] = UNSET
    security: MaybeSet[list[dict[str, list[str]]]] = UNSET
    parameters: MaybeSet[list[Parameter]] = UNSET
    servers: MaybeSet[list[Server]] = UNSET
    component_name_prefix: MaybeSet[str] = UNSET
    errors: MaybeSet[list[type[exceptions.TypedError]]] = UNSET

    def merge(self, other: 'MethodMetadata') -> 'MethodMetadata':
        return MethodMetadata(
            params_schema=self.params_schema if self.params_schema is not UNSET else other.params_schema,
            result_schema=self.result_schema if self.result_schema is not UNSET else other.result_schema,
            summary=self.summary if self.summary is not UNSET else other.summary,
            description=self.description if self.description is not UNSET else other.description,
            external_docs=self.external_docs if self.external_docs is not UNSET else other.external_docs,
            deprecated=self.deprecated if self.deprecated is not UNSET else other.deprecated,
            component_name_prefix=self.component_name_prefix if self.component_name_prefix is not UNSET
            else other.component_name_prefix,
            examples=((self.examples or []) + (other.examples or [])) or UNSET,
            tags=((self.tags or []) + (other.tags or [])) or UNSET,
            security=((self.security or []) + (other.security or [])) or UNSET,
            parameters=((self.parameters or []) + (other.parameters or [])) or UNSET,
            servers=((self.servers or []) + (other.servers or [])) or UNSET,
            errors=((self.errors or []) + (other.errors or [])) or UNSET,
        )


def metadata(
    params_schema: MaybeSet[dict[str, JsonSchema]] = UNSET,
    result_schema: MaybeSet[JsonSchema] = UNSET,
    errors: MaybeSet[list[type[exceptions.TypedError]]] = UNSET,
    examples: MaybeSet[list[MethodExample]] = UNSET,
    tags: MaybeSet[list[Union[str, Tag]]] = UNSET,
    summary: MaybeSet[str] = UNSET,
    description: MaybeSet[str] = UNSET,
    external_docs: MaybeSet[ExternalDocumentation] = UNSET,
    deprecated: MaybeSet[bool] = UNSET,
    security: MaybeSet[list[dict[str, list[str]]]] = UNSET,
    parameters: MaybeSet[list[Parameter]] = UNSET,
    servers: MaybeSet[list[Server]] = UNSET,
    component_name_prefix: MaybeSet[str] = UNSET,
) -> MethodMetadata:
    """
    Adds Open Api specification annotation to the method.

    :param params_schema: method parameters JSON schema
    :param result_schema: method result JSON schema
    :param errors: method errors
    :param examples: method usage examples
    :param tags: a list of tags for method documentation control
    :param summary: a short summary of what the method does
    :param description: a verbose explanation of the method behavior
    :param external_docs: an external resource for extended documentation
    :param deprecated: declares this method to be deprecated
    :param security: a declaration of which security mechanisms can be used for the method
    :param parameters: a list of parameters that are applicable for the method
    :param servers: a list of connectivity information of a target server.
    :param component_name_prefix: components name prefix.
    """

    return MethodMetadata(
        params_schema=params_schema,
        result_schema=result_schema,
        examples=examples,
        tags=[
            tag if isinstance(tag, Tag) else Tag(name=tag) for tag in tags
        ] if not isinstance(tags, UnsetType) else UNSET,
        summary=summary,
        description=description,
        external_docs=external_docs,
        deprecated=deprecated,
        security=security,
        parameters=parameters,
        servers=servers,
        component_name_prefix=component_name_prefix,
        errors=errors,
    )


@dc.dataclass(frozen=True)
class MethodSpecification:
    operation: Operation
    component_schemas: MaybeSet[dict[str, JsonSchema]] = UNSET


class MethodSpecificationGenerator:
    def __init__(self, extractor: BaseMethodInfoExtractor, error_http_status_map: Optional[dict[int, int]] = None):
        self._extractor = extractor
        self._error_http_status_map = error_http_status_map or {}

    def __call__(self, method: Method) -> Method:
        method_metadata = MethodMetadata()
        for meta in method.metadata:
            if isinstance(meta, MethodMetadata):
                method_metadata = method_metadata.merge(meta)

        method_specification = self._generate(method.func, method.name, method_metadata)
        method.metadata.append(method_specification)

        return method

    def _generate(self, method: MethodType, method_name: str, method_metadata: MethodMetadata) -> MethodSpecification:
        status_errors_map = self._extract_errors(method, method_metadata)
        default_status_errors = status_errors_map.pop(HTTP_DEFAULT_STATUS, [])

        component_schemas: dict[str, JsonSchema] = {}

        request_schema, request_components = self._extract_request_schema(method, method_name, method_metadata)
        component_schemas.update(request_components)

        errors_schema, errors_components = self._extract_errors_schema(
            method, method_name, method_metadata, status_errors_map,
        )
        component_schemas.update(errors_components)

        response_schema, response_components = self._extract_response_schema(
            method, method_name, method_metadata, default_status_errors,
        )
        component_schemas.update(response_components)

        request_examples, response_success_examples = self._build_examples(method_name, method_metadata.examples or [])

        tags = method_metadata.tags or []
        servers = method_metadata.servers
        parameters = method_metadata.parameters or []
        security = method_metadata.security
        external_docs = method_metadata.external_docs

        if method_metadata.summary is UNSET:
            summary = self._extractor.extract_summary(method)
        else:
            summary = method_metadata.summary

        if method_metadata.description is UNSET:
            description = self._extractor.extract_description(method)
        else:
            description = method_metadata.description

        if method_metadata.deprecated is UNSET:
            deprecated = self._extractor.extract_deprecation_status(method)
        else:
            deprecated = method_metadata.deprecated

        return MethodSpecification(
            operation=Operation(
                requestBody=RequestBody(
                    description='JSON-RPC Request',
                    content={
                        JSONRPC_MEDIATYPE: MediaType(
                            schema=request_schema,
                            examples=request_examples or UNSET,
                        ),
                    },
                    required=True,
                ),
                responses={
                    **{
                        str(HTTP_DEFAULT_STATUS): Response(
                            description=(response_schema or {}).get('description', 'JSON-RPC Response'),
                            content={
                                JSONRPC_MEDIATYPE: MediaType(
                                    schema=response_schema,
                                    examples=response_success_examples or UNSET,
                                ),
                            },
                        ),
                    },
                    **{
                        str(status): Response(
                            description=error_schema.get('description', 'JSON-RPC Error'),
                            content={
                                JSONRPC_MEDIATYPE: MediaType(
                                    schema=error_schema,
                                ),
                            },
                        )
                        for status, error_schema in errors_schema.items()
                    },
                },
                tags=[tag.name for tag in tags] or UNSET,
                summary=summary,
                description=description,
                deprecated=deprecated,
                externalDocs=external_docs,
                security=security or UNSET,
                parameters=list(parameters) or UNSET,
                servers=servers or UNSET,
            ),
            component_schemas=component_schemas or UNSET,
        )

    def _extract_errors(
        self,
        method: MethodType,
        method_metadata: MethodMetadata,
    ) -> dict[int, list[type[exceptions.TypedError]]]:
        errors = (method_metadata.errors or []) + (self._extractor.extract_errors(method) or [])

        status_error_map: dict[int, list[type[exceptions.TypedError]]] = defaultdict(list)
        unique_errors = list({error.CODE: error for error in errors}.values())
        for error in unique_errors:
            http_status = self._error_http_status_map.get(error.CODE, HTTP_DEFAULT_STATUS)
            status_error_map[http_status].append(error)

        return status_error_map

    def _extract_errors_schema(
        self,
        method: MethodType,
        method_name: str,
        method_metadata: MethodMetadata,
        status_errors_map: dict[int, list[type[exceptions.TypedError]]],
    ) -> tuple[dict[int, dict[str, Any]], dict[str, JsonSchema]]:
        status_error_schema_map: dict[int, dict[str, Any]] = {}
        component_schemas: dict[str, JsonSchema] = {}
        component_name_prefix = method_metadata.component_name_prefix or f"{method.__module__}_"

        for status, errors in status_errors_map.items():
            if result := self._extractor.extract_error_response_schema(
                method_name,
                method,
                ref_template=f'#/components/schemas/{component_name_prefix}{{model}}',
                errors=errors,
            ):
                error_schema, error_component_schemas = result
                if error_component_schemas:
                    component_schemas.update({
                        f"{component_name_prefix}{name}": component
                        for name, component in error_component_schemas.items()
                    })
                status_error_schema_map[status] = error_schema

        return status_error_schema_map, component_schemas

    def _extract_request_schema(
        self,
        method: MethodType,
        method_name: str,
        method_metadata: MethodMetadata,
    ) -> tuple[MaybeSet[dict[str, Any]], dict[str, JsonSchema]]:
        component_schemas: dict[str, JsonSchema] = {}
        component_name_prefix = method_metadata.component_name_prefix or f"{method.__module__}_"

        request_schema: MaybeSet[dict[str, Any]] = UNSET
        if params_schema := method_metadata.params_schema:
            request_schema = build_request_schema(method_name, params_schema)
        else:
            if result := self._extractor.extract_request_schema(
                method_name,
                method,
                ref_template=f'#/components/schemas/{component_name_prefix}{{model}}',
            ):
                request_schema, request_components = result
                if request_components:
                    component_schemas.update({
                        f"{component_name_prefix}{name}": component
                        for name, component in request_components.items()
                    })

        return request_schema, component_schemas

    def _extract_response_schema(
        self,
        method: MethodType,
        method_name: str,
        method_metadata: MethodMetadata,
        errors: list[type[exceptions.TypedError]],
    ) -> tuple[MaybeSet[dict[str, Any]], dict[str, JsonSchema]]:
        component_schemas: dict[str, JsonSchema] = {}
        component_name_prefix = method_metadata.component_name_prefix or f"{method.__module__}_"

        response_schema: MaybeSet[dict[str, Any]] = UNSET
        if result_schema := method_metadata.result_schema:
            response_schema = build_response_schema(result_schema, errors=errors)
        else:
            if result := self._extractor.extract_response_schema(
                method_name,
                method,
                ref_template=f'#/components/schemas/{component_name_prefix}{{model}}',
                errors=errors,
            ):
                response_schema, response_components = result
                if response_components:
                    component_schemas.update({
                        f"{component_name_prefix}{name}": component
                        for name, component in response_components.items()
                    })

        return response_schema, component_schemas

    def _build_examples(self, method_name: str, examples: list[MethodExample]) -> tuple[dict[str, Any], dict[str, Any]]:
        request_examples: dict[str, Any] = {
            example.summary or f'Example#{i}': ExampleObject(
                summary=example.summary,
                description=example.description,
                value={
                    'jsonrpc': example.version,
                    'id': 1,
                    'method': method_name,
                    'params': example.params,
                },
            ) for i, example in enumerate(examples)
        }

        response_examples: dict[str, Any] = {
            example.summary or f'Example#{i}': ExampleObject(
                summary=example.summary,
                description=example.description,
                value={
                    'jsonrpc': example.version,
                    'id': 1,
                    'result': example.result,
                },
            ) for i, example in enumerate(examples)
        }

        return request_examples, response_examples



"""
OpenRPC specification generator. See https://spec.open-rpc.org/.
"""

import dataclasses as dc
from typing import Any, Callable, Iterable, Mapping, Union

from pjrpc.common import UNSET, MaybeSet, UnsetType, exceptions
from pjrpc.server.dispatcher import Method
from pjrpc.server.specs import Specification
from pjrpc.server.specs.extractors import BaseMethodInfoExtractor

from .spec import Components, Contact, ContentDescriptor, Error, ExampleObject, ExternalDocumentation, Info, Json
from .spec import JsonSchema, License, Link, MethodExample, MethodInfo, ParamStructure, Reference, Server
from .spec import ServerVariable, SpecRoot, Tag

__all__ = [
    'Components',
    'Contact',
    'ContentDescriptor',
    'Error',
    'ExampleObject',
    'ExternalDocumentation',
    'Info',
    'Json',
    'JsonSchema',
    'License',
    'Link',
    'metadata',
    'MethodExample',
    'MethodInfo',
    'MethodMetadata',
    'OpenRPC',
    'ParamStructure',
    'Reference',
    'Server',
    'ServerVariable',
    'SpecRoot',
    'Tag',
]

MethodType = Callable[..., Any]


class OpenRPC(Specification):
    """
    OpenRPC Specification.

    :param info: provides metadata about the API
    :param servers: connectivity information
    :param external_docs: additional external documentation
    :param openrpc: the semantic version number of the OpenRPC Specification version that the OpenRPC document uses
    """

    def __init__(
        self,
        info: Info,
        servers: MaybeSet[list[Server]] = UNSET,
        external_docs: MaybeSet[ExternalDocumentation] = UNSET,
        openrpc: str = '1.3.2',
    ):
        self._info = info
        self._servers = servers
        self._external_docs = external_docs
        self._openrpc = openrpc

    def generate(self, root_endpoint: str, methods: Mapping[str, Iterable[Method]]) -> dict[str, Any]:
        spec_root = SpecRoot(
            info=self._info,
            servers=self._servers,
            externalDocs=self._external_docs,
            openrpc=self._openrpc,
            methods=[],
            components=Components(),
        )
        spec_root.components.schemas = {}

        for method in methods.get('', []):
            for meta in method.metadata:
                if isinstance(meta, MethodSpecification):
                    spec_root.methods.append(meta.method_info)
                    spec_root.components.schemas.update(meta.component_schemas or {})

        return dc.asdict(
            spec_root,
            dict_factory=lambda iterable: dict(
                filter(lambda item: not isinstance(item[1], UnsetType), iterable),
            ),
        )


@dc.dataclass(frozen=True)
class MethodMetadata:
    params_schema: MaybeSet[list[ContentDescriptor]] = UNSET
    result_schema: MaybeSet[ContentDescriptor] = UNSET
    errors: MaybeSet[list[Error]] = UNSET
    examples: MaybeSet[list[Union[MethodExample, Reference]]] = UNSET
    tags: MaybeSet[list[Union[Tag, Reference]]] = UNSET
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    deprecated: MaybeSet[bool] = UNSET
    external_docs: MaybeSet[ExternalDocumentation] = UNSET
    servers: MaybeSet[list[Server]] = UNSET

    def merge(self, other: 'MethodMetadata') -> 'MethodMetadata':
        return MethodMetadata(
            params_schema=self.params_schema if self.params_schema is not UNSET else other.params_schema,
            result_schema=self.result_schema if self.result_schema is not UNSET else other.result_schema,
            summary=self.summary if self.summary is not UNSET else other.summary,
            description=self.description if self.description is not UNSET else other.description,
            deprecated=self.deprecated if self.deprecated is not UNSET else other.deprecated,
            external_docs=self.external_docs if self.external_docs is not UNSET else other.external_docs,
            errors=((self.errors or []) + (other.errors or [])) or UNSET,
            examples=((self.examples or []) + (other.examples or [])) or UNSET,
            tags=((self.tags or []) + (other.tags or [])) or UNSET,
            servers=((self.servers or []) + (other.servers or [])) or UNSET,
        )


def metadata(
    params_schema: MaybeSet[list[ContentDescriptor]] = UNSET,
    result_schema: MaybeSet[ContentDescriptor] = UNSET,
    errors: MaybeSet[list[Union[Error, type[exceptions.TypedError]]]] = UNSET,
    examples: MaybeSet[list[Union[MethodExample, Reference]]] = UNSET,
    summary: MaybeSet[str] = UNSET,
    description: MaybeSet[str] = UNSET,
    tags: MaybeSet[list[Union[Tag, str]]] = UNSET,
    deprecated: MaybeSet[bool] = UNSET,
    external_docs: MaybeSet[ExternalDocumentation] = UNSET,
    servers: MaybeSet[list[Server]] = UNSET,
) -> MethodMetadata:
    """
    Adds JSON-RPC method to the API specification.

    :param params_schema: a list of parameters that are applicable for this method
    :param result_schema: the description of the result returned by the method
    :param errors: a list of custom application defined errors that MAY be returned
    :param examples: method usage example
    :param summary: a short summary of what the method does
    :param description: a verbose explanation of the method behavior
    :param tags: a list of tags for API documentation control
    :param deprecated: declares this method to be deprecated
    :param external_docs: additional external documentation for the method
    :param servers: an alternative servers array to service this method
    """

    return MethodMetadata(
        params_schema=params_schema,
        result_schema=result_schema,
        errors=[
            error if isinstance(error, Error) else Error(code=error.CODE, message=error.MESSAGE)
            for error in errors
        ] if errors else UNSET,
        examples=examples,
        tags=[
            Tag(name=tag) if isinstance(tag, str) else tag
            for tag in tags
        ] if tags else UNSET,
        summary=summary,
        description=description,
        deprecated=deprecated,
        external_docs=external_docs,
        servers=servers,
    )


@dc.dataclass(frozen=True)
class MethodSpecification:
    method_info: MethodInfo
    component_schemas: MaybeSet[dict[str, JsonSchema]] = UNSET


class MethodSpecificationGenerator:
    def __init__(self, extractor: BaseMethodInfoExtractor):
        self._extractor = extractor

    def __call__(self, method: Method) -> Method:
        method_metadata = MethodMetadata()
        for meta in method.metadata:
            if isinstance(meta, MethodMetadata):
                method_metadata = method_metadata.merge(meta)

        method_specification = self._generate(method.func, method.name, method_metadata)
        method.metadata.append(method_specification)

        return method

    def _generate(self, method: MethodType, method_name: str, method_metadata: MethodMetadata) -> MethodSpecification:
        tags = method_metadata.tags
        examples = method_metadata.examples
        servers = method_metadata.servers
        external_docs = method_metadata.external_docs

        if method_metadata.summary is UNSET:
            summary = self._extractor.extract_summary(method)
        else:
            summary = method_metadata.summary

        if method_metadata.description is UNSET:
            description = self._extractor.extract_description(method)
        else:
            description = method_metadata.description

        if method_metadata.deprecated is UNSET:
            deprecated = self._extractor.extract_deprecation_status(method)
        else:
            deprecated = method_metadata.deprecated

        component_schemas: dict[str, JsonSchema] = {}
        params_schema, params_component_schemas = self._extract_params_schema(method, method_name, method_metadata)
        component_schemas.update(params_component_schemas)

        result_schema, result_component_schemas = self._extract_result_schema(method, method_name, method_metadata)
        component_schemas.update(result_component_schemas)

        errors = self._extract_errors(method, method_metadata)

        method_info = MethodInfo(
            name=method_name,
            params=list(params_schema),
            result=result_schema,
            errors=list(errors) if errors else UNSET,
            examples=examples,
            summary=summary,
            description=description,
            tags=tags,
            deprecated=deprecated,
            externalDocs=external_docs,
            servers=servers,
        )

        return MethodSpecification(method_info, component_schemas)

    def _extract_params_schema(
        self,
        method: MethodType,
        method_name: str,
        method_metadata: MethodMetadata,
    ) -> tuple[list[ContentDescriptor], dict[str, JsonSchema]]:
        component_schemas: dict[str, JsonSchema] = {}
        if not (params_descriptors := method_metadata.params_schema):
            request_ref_prefix = '#/components/schemas/'
            params_schema, component_schemas = self._extractor.extract_params_schema(
                method_name,
                method,
                ref_template=f'{request_ref_prefix}{{model}}',
            )
            params_descriptors = [
                ContentDescriptor(
                    name=name,
                    schema=schema,
                    summary=schema.get('title', UNSET),
                    description=schema.get('description', UNSET),
                    required=name in params_schema.get('required', []),
                    deprecated=schema.get('deprecated', UNSET),
                ) for name, schema in params_schema['properties'].items()
            ]

        return params_descriptors, component_schemas

    def _extract_result_schema(
        self,
        method: MethodType,
        method_name: str,
        method_metadata: MethodMetadata,
    ) -> tuple[ContentDescriptor, dict[str, JsonSchema]]:
        component_schemas: dict[str, JsonSchema] = {}
        if not (result_descriptor := method_metadata.result_schema):
            response_ref_prefix = '#/components/schemas/'
            result_schema, component_schemas = self._extractor.extract_result_schema(
                method_name,
                method,
                ref_template=f'{response_ref_prefix}{{model}}',
            )
            result_descriptor = ContentDescriptor(
                name='result',
                schema=result_schema,
                summary=result_schema.get('title', UNSET),
                description=result_schema.get('description', UNSET),
                required='result' in result_schema.get('required', []),
                deprecated=result_schema.get('deprecated', UNSET),
            )

        return result_descriptor, component_schemas

    def _extract_errors(self, method: MethodType, method_metadata: MethodMetadata) -> MaybeSet[list[Error]]:
        errors = method_metadata.errors or []
        errors.extend([
            Error(code=error.CODE, message=error.MESSAGE)
            for error in self._extractor.extract_errors(method) or []
        ])

        unique_errors = list({error.code: error for error in errors}.values())

        return unique_errors or UNSET



"""
JSON-RPC version 2.0 protocol request implementation.
"""

import abc
import dataclasses as dc
import itertools as it
import operator as op
from typing import Any, ClassVar, Iterator, Optional

from pjrpc.common.typedefs import JsonRpcParamsT, JsonRpcRequestIdT

from .common import JsonT
from .exceptions import DeserializationError, IdentityError


@dc.dataclass
class AbstractRequest(abc.ABC):
    """
    JSON-RPC 2.0 abstract request.
    """

    @classmethod
    @abc.abstractmethod
    def from_json(cls, data: JsonT) -> 'AbstractRequest':
        pass

    @abc.abstractmethod
    def to_json(self) -> JsonT:
        pass

    @property
    @abc.abstractmethod
    def is_notification(self) -> bool:
        pass


@dc.dataclass
class Request(AbstractRequest):
    """
    JSON-RPC 2.0 request.

    :param method: method name
    :param params: method parameters
    :param id: request identifier
    """

    version: ClassVar[str] = '2.0'

    method: str
    params: Optional[JsonRpcParamsT] = None
    id: Optional[JsonRpcRequestIdT] = None

    @classmethod
    def from_json(cls, data: JsonT) -> 'Request':
        """
        Deserializes a request from json data.

        :param data: data the request to be deserialized from
        :returns: request object
        :raises: :py:class:`pjrpc.common.exceptions.DeserializationError` if format is incorrect
        """

        try:
            if not isinstance(data, dict):
                raise DeserializationError("data must be of type dict")

            jsonrpc = data['jsonrpc']
            if not isinstance(jsonrpc, str):
                raise DeserializationError("field 'jsonrpc' must be of type string")

            if jsonrpc != cls.version:
                raise DeserializationError(f"jsonrpc version '{data['jsonrpc']}' is not supported")

            request_id = data.get('id')
            if request_id is not None and not isinstance(request_id, (int, str)):
                raise DeserializationError("field 'id' must be of type integer or string")

            method = data['method']
            if not isinstance(method, str):
                raise DeserializationError("field 'method' must be of type string")

            params = data.get('params', {})
            if not isinstance(params, (list, dict)):
                raise DeserializationError("field 'params' must be of type list or dict")

            return cls(id=request_id, method=method, params=params)
        except KeyError as e:
            raise DeserializationError(f"required field {e} not found") from e

    def to_json(self) -> JsonT:
        """
        Serializes the request to json data.

        :returns: json data
        """

        data: dict[str, JsonT] = {
            'jsonrpc': self.version,
            'method': self.method,
        }
        if self.id is not None:
            data.update(id=self.id)
        if self.params:
            data.update(params=self.params)

        return data

    @property
    def is_notification(self) -> bool:
        """
        Returns ``True`` if the request is a notification e.g. response will not be sent.
        """

        return self.id is None


@dc.dataclass
class BatchRequest(AbstractRequest):
    """
    JSON-RPC 2.0 batch request.

    :param requests: requests to be added to the batch
    """

    version: ClassVar[str] = '2.0'

    requests: tuple[Request, ...]

    def __init__(self, *requests: Request) -> None:
        self.requests = tuple(requests)

    @classmethod
    def from_json(cls, data: JsonT, check_ids: bool = True) -> 'BatchRequest':
        """
        Deserializes a batch request from json data.

        :param data: data the request to be deserialized from
        :param check_ids: check response ids for uniqueness
        :returns: batch request object
        """

        if not isinstance(data, (list, tuple)):
            raise DeserializationError("data must be of type list")

        if len(data) == 0:
            raise DeserializationError("request list is empty")

        requests = tuple(Request.from_json(request) for request in data)
        if check_ids:
            cls._check_response_id_uniqueness(requests)

        return cls(*requests)

    def to_json(self) -> JsonT:
        """
        Serializes the request to json data.

        :returns: json data
        """

        return [request.to_json() for request in self.requests]

    def __getitem__(self, idx: int) -> Request:
        """
        Returns a request at index `idx`.
        """

        return self.requests[idx]

    def __iter__(self) -> Iterator[Request]:
        return iter(self.requests)

    def __len__(self) -> int:
        return len(self.requests)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, BatchRequest):
            return NotImplemented

        return all(
            it.starmap(
                op.eq, zip(
                    sorted(self, key=op.attrgetter('id')),
                    sorted(other, key=op.attrgetter('id')),
                ),
            ),
        )

    @property
    def is_notification(self) -> bool:
        """
        Returns ``True`` if all requests in the batch are notifications.
        """

        return all(map(op.attrgetter('is_notification'), self.requests))

    @classmethod
    def _check_response_id_uniqueness(cls, requests: tuple[Request, ...]) -> None:
        ids = tuple(request.id for request in requests if request.id is not None)
        if len(ids) != len(set(ids)):
            raise IdentityError("batch request ids are not unique")



"""
Definition of package exceptions and JSON-RPC protocol errors.
"""

import dataclasses as dc
from typing import Any, ClassVar, Optional

from .common import UNSET, MaybeSet

JsonT = Any


class BaseError(Exception):
    """
    Base package error. All package errors are inherited from it.
    """


class ProtocolError(BaseError):
    """
    Raised when JSON-RPC protocol is violated.
    """


class IdentityError(ProtocolError):
    """
    Raised when a batch requests/responses identifiers are not unique or missing.
    """


class DeserializationError(ProtocolError, ValueError):
    """
    Request/response deserializatoin error.
    Raised when request/response json has incorrect format.
    """


@dc.dataclass(frozen=True)
class JsonRpcError(BaseError):
    """
    `JSON-RPC <https://www.jsonrpc.org>`_ protocol error.
    For more information see `Error object <https://www.jsonrpc.org/specification#error_object>`_.
    All JSON-RPC protocol errors are inherited from it.

    :param code: number that indicates the error type
    :param message: short description of the error
    :param data: value that contains additional information about the error. May be omitted.
    """

    # typed subclasses error mapping
    __TYPED_ERRORS__: ClassVar[dict[int, type['TypedError']]] = {}

    code: int
    message: str
    data: MaybeSet[JsonT] = dc.field(repr=False, default=UNSET)

    @classmethod
    def get_error_by_code(cls, code: int) -> Optional[type['TypedError']]:
        return cls.__TYPED_ERRORS__.get(code)

    @classmethod
    def from_json(cls, json_data: JsonT) -> 'JsonRpcError':
        """
        Deserializes an error from json data. If data format is not correct :py:class:`ValueError` is raised.

        :param json_data: json data the error to be deserialized from

        :returns: deserialized error
        :raises: :py:class:`pjrpc.common.exceptions.DeserializationError` if format is incorrect
        """

        try:
            if not isinstance(json_data, dict):
                raise DeserializationError("data must be of type dict")

            code = json_data['code']
            if not isinstance(code, int):
                raise DeserializationError("field 'code' must be of type integer")

            message = json_data['message']
            if not isinstance(message, str):
                raise DeserializationError("field 'message' must be of type string")

            data = json_data.get('data', UNSET)

            if error_class := cls.get_error_by_code(code):
                return error_class(message, data)
            else:
                return JsonRpcError(code, message, data)

        except KeyError as e:
            raise DeserializationError(f"required field {e} not found") from e

    def to_json(self) -> JsonT:
        """
        Serializes the error into a dict.

        :returns: serialized error
        """

        json: dict[str, JsonT] = {
            'code': self.code,
            'message': self.message,
        }
        if self.data is not UNSET:
            json.update(data=self.data)

        return json

    def __str__(self) -> str:
        return f"({self.code}) {self.message}"


class TypedError(JsonRpcError):
    """
    Typed JSON-RPC error.
    Must not be instantiated directly, only subclassed.
    """

    # a number that indicates the error type that occurred
    CODE: ClassVar[int]

    # a string providing a short description of the error.
    # the message SHOULD be limited to a concise single sentence.
    MESSAGE: ClassVar[str]

    def __init_subclass__(cls, base: bool = False, **kwargs: Any):
        super().__init_subclass__(**kwargs)
        if base:
            cls.__TYPED_ERRORS__ = cls.__TYPED_ERRORS__.copy()

        if issubclass(cls, TypedError) and (code := getattr(cls, 'CODE', None)) is not None:
            cls.__TYPED_ERRORS__[code] = cls

    def __init__(self, message: Optional[str] = None, data: MaybeSet[JsonT] = UNSET):
        super().__init__(self.CODE, message or self.MESSAGE, data)


class ClientBaseError(TypedError):
    """
    Raised when a client sent an incorrect request.
    Must not be instantiated directly, only subclassed.
    """


class ParseError(ClientBaseError):
    """
    Invalid JSON was received by the server.
    An error occurred on the server while parsing the JSON text.
    """

    CODE: ClassVar[int] = -32700
    MESSAGE: ClassVar[str] = 'Parse error'


class InvalidRequestError(ClientBaseError):
    """
    The JSON sent is not a valid request object.
    """

    CODE: ClassVar[int] = -32600
    MESSAGE: ClassVar[str] = 'Invalid Request'


class MethodNotFoundError(ClientBaseError):
    """
    The method does not exist / is not available.
    """

    CODE: ClassVar[int] = -32601
    MESSAGE: ClassVar[str] = 'Method not found'


class InvalidParamsError(ClientBaseError):
    """
    Invalid method parameter(s).
    """

    CODE: ClassVar[int] = -32602
    MESSAGE: ClassVar[str] = 'Invalid params'


class ServerBaseError(TypedError):
    """
    Raised when a server failed to process a reqeust.
    Must not be instantiated directly, only subclassed.
    """


class InternalError(ServerBaseError):
    """
    Internal JSON-RPC error.
    """

    CODE: ClassVar[int] = -32603
    MESSAGE: ClassVar[str] = 'Internal error'


class ServerError(ServerBaseError):
    """
    Reserved for implementation-defined server-errors.
    Codes from -32000 to -32099.
    """

    CODE: ClassVar[int] = -32000
    MESSAGE: ClassVar[str] = 'Server error'



import json
import typing
from ssl import SSLContext
from typing import Any, Callable, Generator, Iterable, Mapping, Optional, TypedDict, Union

from aiohttp import BasicAuth, Fingerprint, client
from aiohttp.typedefs import LooseCookies, LooseHeaders, StrOrURL
from multidict import MultiDict

import pjrpc
from pjrpc.client import AbstractAsyncClient, AsyncMiddleware
from pjrpc.common import AbstractRequest, AbstractResponse, BatchRequest, BatchResponse, JSONEncoder, JsonRpcError
from pjrpc.common import Request, Response, generators
from pjrpc.common.typedefs import JsonRpcRequestIdT


class RequestArgs(TypedDict, total=False):
    cookies: LooseCookies
    headers: LooseHeaders
    skip_auto_headers: Iterable[str]
    auth: BasicAuth
    allow_redirects: bool
    max_redirects: int
    compress: Union[str, bool]
    chunked: bool
    expect100: bool
    read_until_eof: bool
    proxy: StrOrURL
    proxy_auth: BasicAuth
    timeout: "Union[client.ClientTimeout, None]"
    ssl: Union[SSLContext, bool, Fingerprint]
    server_hostname: str
    proxy_headers: LooseHeaders
    trace_request_ctx: Mapping[str, Any]
    read_bufsize: int
    auto_decompress: bool
    max_line_size: int
    max_field_size: int


class Client(AbstractAsyncClient):
    """
    `Aiohttp <https://aiohttp.readthedocs.io/en/stable/client.html>`_ library client backend.

    :param url: url to be used as JSON-RPC endpoint
    :param session: custom session to be used instead of :py:class:`aiohttp.ClientSession`
    :param raise_for_status: should `ClientResponse.raise_for_status()` be called automatically
    :param id_gen_impl: identifier generator
    :param error_cls: JSON-RPC error base class
    :param json_loader: json loader
    :param json_dumper: json dumper
    :param json_encoder: json encoder
    :param json_decoder: json decoder
    """

    def __init__(
        self,
        url: str,
        *,
        session: Optional[client.ClientSession] = None,
        raise_for_status: bool = True,
        id_gen_impl: Callable[..., Generator[JsonRpcRequestIdT, None, None]] = generators.sequential,
        error_cls: type[JsonRpcError] = JsonRpcError,
        json_loader: Callable[..., Any] = json.loads,
        json_dumper: Callable[..., str] = json.dumps,
        json_encoder: type[JSONEncoder] = JSONEncoder,
        json_decoder: Optional[json.JSONDecoder] = None,
        middlewares: Iterable[AsyncMiddleware] = (),
    ):
        super().__init__(
            id_gen_impl=id_gen_impl,
            error_cls=error_cls,
            json_loader=json_loader,
            json_dumper=json_dumper,
            json_encoder=json_encoder,
            json_decoder=json_decoder,
            middlewares=middlewares,
        )
        self._endpoint = url
        self._session = session or client.ClientSession()
        self._owned_session = session is None
        self._raise_for_status = raise_for_status

    @typing.overload
    async def send(self, request: Request, **kwargs: Any) -> Optional[Response]:
        ...

    @typing.overload
    async def send(self, request: BatchRequest, **kwargs: Any) -> Optional[BatchResponse]:
        ...

    async def send(self, request: AbstractRequest, **kwargs: Any) -> Optional[AbstractResponse]:
        """
        Sends a JSON-RPC request.

        :param request: request instance
        :param kwargs: additional client request argument
        :returns: response instance or None if the request is a notification
        """

        return await self._send(request, kwargs)

    async def _request(
            self,
            request_text: str,
            is_notification: bool,
            request_kwargs: Mapping[str, Any],
    ) -> Optional[str]:
        """
        Makes a JSON-RPC request.

        :param request_text: request text representation
        :param is_notification: is the request a notification
        :param request_kwargs: additional client request argument
        :returns: response text representation or None if the request is a notification
        """

        request_kwargs = typing.cast(RequestArgs, request_kwargs)

        request_kwargs['headers'] = headers = MultiDict(request_kwargs.get('headers', {}))
        headers['Content-Type'] = self._request_content_type

        async with self._session.post(self._endpoint, data=request_text, **request_kwargs) as resp:
            if self._raise_for_status:
                resp.raise_for_status()
            response_text = await resp.text()

        if is_notification:
            return None

        content_type = resp.headers.get('Content-Type', '')
        if response_text and content_type.split(';')[0] not in self._response_content_types:
            raise pjrpc.exc.DeserializationError(f"unexpected response content type: {content_type}")

        return response_text

    async def close(self) -> None:
        """
        Closes current http session.
        """

        if self._owned_session:
            await self._session.close()

    async def __aenter__(self) -> 'Client':
        if self._owned_session:
            await self._session.__aenter__()

        return self

    async def __aexit__(self, *args: Any) -> None:
        if self._owned_session:
            await self._session.__aexit__(*args)



"""
OpenAPI Specification. See https://swagger.io/specification/.
"""

import dataclasses as dc
import enum
from typing import Any, Union

from pjrpc.common import UNSET, MaybeSet

Json = Union[str, int, float, dict[str, 'Json'], bool, list['Json'], tuple['Json'], set['Json'], None]
JsonSchema = dict[str, Json]


@dc.dataclass
class Reference:
    """
    A simple object to allow referencing other components in the OpenAPI document, internally and externally.

    :param ref: the reference identifier.
    :param summary: a short summary which by default SHOULD override that of the referenced component.
    :param description: a description which by default SHOULD override that of the referenced component
    """

    ref: str
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET

    def __post_init__(self) -> None:
        self.__dict__['$ref'] = self.__dict__['ref']
        self.__dataclass_fields__['ref'].name = '$ref'  # noqa


@dc.dataclass
class Contact:
    """
    Contact information for the exposed API.

    :param name: the identifying name of the contact person/organization
    :param url: the URL pointing to the contact information
    :param email: the email address of the contact person/organization
    """

    name: MaybeSet[str] = UNSET
    url: MaybeSet[str] = UNSET
    email: MaybeSet[str] = UNSET


@dc.dataclass
class License:
    """
    License information for the exposed API.

    :param name: the license name used for the API
    :param identifier: an SPDX license expression for the API
    :param url: a URL to the license used for the API
    """

    name: str
    identifier: MaybeSet[str] = UNSET
    url: MaybeSet[str] = UNSET


@dc.dataclass
class Info:
    """
    Metadata about the API.

    :param title: the title of the application
    :param version: the version of the OpenAPI document
    :param summary: a short summary of the API.
    :param description: a description of the application
    :param contact: the contact information for the exposed API
    :param license: the license information for the exposed API
    :param termsOfService: a URL to the Terms of Service for the API
    """

    title: str
    version: str
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    contact: MaybeSet[Contact] = UNSET
    license: MaybeSet[License] = UNSET
    termsOfService: MaybeSet[str] = UNSET


@dc.dataclass
class ServerVariable:
    """
    An object representing a Server Variable for server URL template substitution.

    :param default: the default value to use for substitution, which SHALL be sent if an alternate value is not supplied
    :param enum: an enumeration of string values to be used if the substitution options are from a limited set
    :param description: an optional description for the server variable
    """

    default: str
    enum: MaybeSet[list[str]] = UNSET
    description: MaybeSet[str] = UNSET


@dc.dataclass
class Server:
    """
    Connectivity information of a target server.

    :param url: a URL to the target host
    :param description: an optional string describing the host designated by the URL
    :param variables: a map between a variable name and its value.
                      The value is used for substitution in the server's URL template.
    """

    url: str
    description: MaybeSet[str] = UNSET
    variables: MaybeSet[dict[str, ServerVariable]] = UNSET


@dc.dataclass
class Link:
    """
    The Link object represents a possible design-time link for a response.

    :param operationRef: a relative or absolute URI reference to an OAS operation
    :param operationId: the name of an existing, resolvable OAS operation, as defined with a unique operationId
    :param parameters: a map representing parameters to pass to an operation as specified with operationId
                       or identified via operationRef
    :param requestBody: a literal value or {expression} to use as a request body when calling the target operation
    :param description: a description of the link
    :param server: a server object to be used by the target operation.
    """

    operationRef: MaybeSet[str] = UNSET
    operationId: MaybeSet[str] = UNSET
    parameters: MaybeSet[dict[str, Any]] = UNSET
    requestBody: MaybeSet[Any] = UNSET
    description: MaybeSet[str] = UNSET
    server: MaybeSet[Server] = UNSET


@dc.dataclass
class ExternalDocumentation:
    """
    Allows referencing an external resource for extended documentation.

    :param url: a short description of the target documentation.
    :param description: the URL for the target documentation
    """

    url: str
    description: MaybeSet[str] = UNSET


@dc.dataclass
class Tag:
    """
    A list of tags for API documentation control.
    Tags can be used for logical grouping of methods by resources or any other qualifier.

    :param name: the name of the tag
    :param externalDocs: additional external documentation for this tag
    :param description: a short description for the tag
    """

    name: str
    description: MaybeSet[str] = UNSET
    externalDocs: MaybeSet[ExternalDocumentation] = UNSET


class SecuritySchemeType(str, enum.Enum):
    """
    The type of the security scheme.
    """

    APIKEY = 'apiKey'
    HTTP = 'http'
    OAUTH2 = 'oauth2'
    OPENID_CONNECT = 'openIdConnect'


class ApiKeyLocation(str, enum.Enum):
    """
    The location of the API key.
    """

    QUERY = 'query'
    HEADER = 'header'
    COOKIE = 'cookie'


@dc.dataclass
class OAuthFlow:
    """
    Configuration details for a supported OAuth Flow.

    :param authorizationUrl: the authorization URL to be used for this flow
    :param tokenUrl: the token URL to be used for this flow
    :param refreshUrl: the URL to be used for obtaining refresh tokens
    :param scopes: the available scopes for the OAuth2 security scheme
    """

    authorizationUrl: str
    tokenUrl: str
    scopes: dict[str, str]
    refreshUrl: MaybeSet[str] = UNSET


@dc.dataclass
class OAuthFlows:
    """
    Configuration of the supported OAuth Flows.

    :param implicit: configuration for the OAuth Implicit flow
    :param password: configuration for the OAuth Resource Owner Password flow
    :param clientCredentials: configuration for the OAuth Client Credentials flow
    :param authorizationCode: configuration for the OAuth Authorization Code flow
    """

    implicit: MaybeSet[OAuthFlow] = UNSET
    password: MaybeSet[OAuthFlow] = UNSET
    clientCredentials: MaybeSet[OAuthFlow] = UNSET
    authorizationCode: MaybeSet[OAuthFlow] = UNSET


@dc.dataclass
class SecurityScheme:
    """
    Defines a security scheme that can be used by the operations.

    :param type: the type of the security scheme
    :param name: the name of the header, query or cookie parameter to be used
    :param location: the location of the API key
    :param scheme: the name of the HTTP Authorization scheme to be used in the Authorization header
    :param bearerFormat: a hint to the client to identify how the bearer token is formatted
    :param flows: an object containing configuration information for the flow types supported
    :param openIdConnectUrl:
    :param description: a short description for security scheme
    """

    type: SecuritySchemeType
    scheme: MaybeSet[str] = UNSET
    name: MaybeSet[str] = UNSET
    location: MaybeSet[ApiKeyLocation] = UNSET  # `in` field
    bearerFormat: MaybeSet[str] = UNSET
    flows: MaybeSet[OAuthFlows] = UNSET
    openIdConnectUrl: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET

    def __post_init__(self) -> None:
        # `in` field name is not allowed in python
        self.__dict__['in'] = self.__dict__['location']
        self.__dataclass_fields__['location'].name = 'in'  # noqa


@dc.dataclass
class MethodExample:
    """
    Method usage example.

    :param params: example parameters
    :param result: example result
    :param summary: short description for the example pairing
    :param description: a verbose explanation of the example pairing
    """

    params: dict[str, Any]
    result: Any
    version: str = '2.0'
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET


@dc.dataclass
class ExampleObject:
    """
    Method usage example.

    :param value: embedded literal example
    :param summary: short description for the example.
    :param description: long description for the example
    :param externalValue: a URL that points to the literal example
    """

    value: MaybeSet[Any] = UNSET
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    externalValue: MaybeSet[str] = UNSET


@dc.dataclass
class Encoding:
    """
    A single encoding definition applied to a single schema property.

    :param contentType: the Content-Type for encoding a specific property
    :param headers: a map allowing additional information to be provided as headers
    :param style: describes how a specific property value will be serialized depending on its type
    :param explode: when this is true, property values of type array or object generate separate parameters
                    for each value of the array, or key-value-pair of the map
    :param allowReserved: determines whether the parameter value SHOULD allow reserved characters,
                          as defined by RFC3986 to be included without percent-encoding
    """

    contentType: MaybeSet[str] = UNSET
    headers: MaybeSet[dict[str, Union['Header', Reference]]] = UNSET
    style: MaybeSet[str] = UNSET
    explode: MaybeSet[bool] = UNSET
    allowReserved: MaybeSet[bool] = UNSET


class ParameterLocation(str, enum.Enum):
    """
    The location of the parameter.
    """

    QUERY = 'query'
    HEADER = 'header'
    PATH = 'path'
    COOKIE = 'cookie'


class StyleType(str, enum.Enum):
    """
    Describes how the parameter value will be serialized depending on the type of the parameter value.
    """

    MATRIX = 'matrix'  # path-style parameters defined by RFC6570
    LABEL = 'label'  # label style parameters defined by RFC6570
    FORM = 'form'  # form style parameters defined by RFC6570
    SIMPLE = 'simple'  # simple style parameters defined by RFC6570
    SPACE_DELIMITED = 'spaceDelimited'  # space separated array values
    PIPE_DELIMITED = 'pipeDelimited'  # pipe separated array values
    DEEP_OBJECT = 'deepObject'  # provides a simple way of rendering nested objects using form parameters


@dc.dataclass
class MediaType:
    """
    Each Media Type Object provides schema and examples for the media type identified by its key.

    :param schema: the schema defining the content.
    :param example: example of the media type
    """

    schema: MaybeSet[dict[str, Any]] = UNSET
    example: MaybeSet[Any] = UNSET
    examples: MaybeSet[dict[str, ExampleObject]] = UNSET
    encoding: MaybeSet[dict[str, Encoding]] = UNSET


@dc.dataclass
class RequestBody:
    """
    Describes a single request body.

    :param content: the content of the request body
    :param required: determines if the request body is required in the request
    :param description: a brief description of the request body
    """

    content: dict[str, MediaType]
    required: MaybeSet[bool] = UNSET
    description: MaybeSet[str] = UNSET


@dc.dataclass
class Parameter:
    """
    Describes a single operation parameter.

    :param name: the name of the parameter
    :param location: the location of the parameter
    :param description: a brief description of the parameter
    :param required: determines whether this parameter is mandatory
    :param deprecated: a parameter is deprecated and SHOULD be transitioned out of usage
    :param allowEmptyValue: the ability to pass empty-valued parameters
    :param style: describes how the parameter value will be serialized depending on the type of the parameter value
    :param explode: when this is true, parameter values of type array or object generate separate parameters
                    for each value of the array or key-value pair of the map
    :param allowReserved: determines whether the parameter value SHOULD allow reserved characters,
                          as defined by RFC3986 :/?#[]@!$&'()*+,;= to be included without percent-encoding
    :param schema: the schema defining the type used for the parameter.
    :param examples: examples of the parameter's potential value
    :param content: a map containing the representations for the parameter
    """

    name: str
    location: ParameterLocation  # `in` field
    description: MaybeSet[str] = UNSET
    required: MaybeSet[bool] = UNSET
    deprecated: MaybeSet[bool] = UNSET
    allowEmptyValue: MaybeSet[bool] = UNSET
    style: MaybeSet[StyleType] = UNSET
    explode: MaybeSet[bool] = UNSET
    allowReserved: MaybeSet[bool] = UNSET
    schema: MaybeSet[JsonSchema] = UNSET
    example: MaybeSet[Any] = UNSET
    examples:  MaybeSet[dict[str, ExampleObject]] = UNSET
    content: MaybeSet[dict[str, MediaType]] = UNSET

    def __post_init__(self) -> None:
        # `in` field name is not allowed in python
        self.__dict__['in'] = self.__dict__['location']
        self.__dataclass_fields__['location'].name = 'in'  # noqa


Header = Parameter


@dc.dataclass
class Response:
    """
    A container for the expected responses of an operation.

    :param description: a short description of the response
    :param content: a map containing descriptions of potential response payloads
    """

    description: str
    headers: MaybeSet[dict[str, Union[Header, Reference]]] = UNSET
    content: MaybeSet[dict[str, MediaType]] = UNSET
    links: MaybeSet[dict[str, Union[Link, Reference]]] = UNSET


@dc.dataclass
class Operation:
    """
    Describes a single API operation on a path.

    :param tags: a list of tags for API documentation control
    :param summary: a short summary of what the operation does
    :param description: a verbose explanation of the operation behavior
    :param externalDocs: additional external documentation for this operation
    :param requestBody: the request body applicable for this operation
    :param responses: the list of possible responses as they are returned from executing this operation
    :param deprecated: declares this operation to be deprecated
    :param servers: an alternative server array to service this operation
    :param security: a declaration of which security mechanisms can be used for this operation
    """

    responses: MaybeSet[dict[str, Union[Response, Reference]]] = UNSET
    requestBody: MaybeSet[Union[RequestBody, Reference]] = UNSET
    tags: MaybeSet[list[str]] = UNSET
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    externalDocs: MaybeSet[ExternalDocumentation] = UNSET
    operationId: MaybeSet[str] = UNSET
    deprecated: MaybeSet[bool] = UNSET
    servers: MaybeSet[list[Server]] = UNSET
    security: MaybeSet[list[dict[str, list[str]]]] = UNSET
    parameters: MaybeSet[list[Union[Parameter, Reference]]] = UNSET


@dc.dataclass
class Path:
    """
    Describes the interface for the given method name.

    :param summary: an optional, string summary, intended to apply to all operations in this path
    :param description: an optional, string description, intended to apply to all operations in this path
    :param servers: an alternative server array to service all operations in this path
    :param parameters: a list of parameters that are applicable for all the operations described under this path
    """

    get: MaybeSet[Operation] = UNSET
    put: MaybeSet[Operation] = UNSET
    post: MaybeSet[Operation] = UNSET
    delete: MaybeSet[Operation] = UNSET
    options: MaybeSet[Operation] = UNSET
    head: MaybeSet[Operation] = UNSET
    patch: MaybeSet[Operation] = UNSET
    trace: MaybeSet[Operation] = UNSET
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    servers: MaybeSet[list[Server]] = UNSET
    parameters: MaybeSet[Union[Parameter, Reference]] = UNSET


@dc.dataclass
class Components:
    """
    Holds a set of reusable objects for different aspects of the OAS.

    :param securitySchemes: an object to hold reusable Security Scheme Objects
    :param schemas: the definition of input and output data types
    """

    schemas: MaybeSet[dict[str, JsonSchema]] = UNSET
    responses: MaybeSet[dict[str, Union[Response, Reference]]] = UNSET
    parameters: MaybeSet[dict[str, Union[Parameter, Reference]]] = UNSET
    examples: MaybeSet[dict[str, Union[ExampleObject, Reference]]] = UNSET
    requestBodies: MaybeSet[dict[str, Union[RequestBody, Reference]]] = UNSET
    headers: MaybeSet[dict[str, Union[Header, Reference]]] = UNSET
    securitySchemes: MaybeSet[dict[str, Union[SecurityScheme, Reference]]] = UNSET
    links: MaybeSet[dict[str, dict[str, Union[Link, Reference]]]] = UNSET
    pathItems: MaybeSet[dict[str, Union[Path, Reference]]] = UNSET


@dc.dataclass
class SpecRoot:
    """
    The root object of the OpenAPI description.

    :param info: provides metadata about the API
    :param servers: an array of Server Objects, which provide connectivity information to a target server
    :param externalDocs: additional external documentation
    :param openapi: the semantic version number of the OpenAPI Specification version that the OpenAPI document uses
    :param jsonSchemaDialect: the default value for the $schema keyword within Schema Objects
    :param tags: a list of tags used by the specification with additional metadata
    :param security: a declaration of which security mechanisms can be used across the API
    """

    info: Info
    paths: dict[str, Path]
    components: Components
    servers: MaybeSet[list[Server]] = UNSET
    externalDocs: MaybeSet[ExternalDocumentation] = UNSET
    tags: MaybeSet[list[Tag]] = UNSET
    security: MaybeSet[list[dict[str, list[str]]]] = UNSET
    openapi: str = '3.1.0'
    jsonSchemaDialect: MaybeSet[str] = UNSET



import asyncio
import dataclasses as dc
import itertools as it
import logging
import time
from typing import Any, Callable, Generator, Iterator, Mapping, Optional

from pjrpc.client import AsyncMiddlewareHandler, MiddlewareHandler
from pjrpc.common import AbstractRequest, AbstractResponse

logger = logging.getLogger(__package__)

Jitter = Callable[[int], float]


@dc.dataclass(frozen=True)
class Backoff:
    """
    JSON-RPC request retry strategy.

    :param attempts: retries number
    :param jitter: retry delay jitter generator
    """

    attempts: int
    jitter: Jitter = lambda attempt: 0.0

    def __call__(self) -> Iterator[float]:
        """
        Returns delay iterator.
        """

        raise NotImplementedError


@dc.dataclass(frozen=True)
class PeriodicBackoff(Backoff):
    """
    Periodic request retry strategy.

    :param interval: retry delay
    """

    interval: float = 1.0

    def __call__(self) -> Iterator[float]:
        def gen() -> Generator[float, None, None]:
            for attempt in range(self.attempts):
                yield self.interval + self.jitter(attempt)

        return gen()


@dc.dataclass(frozen=True)
class ExponentialBackoff(Backoff):
    """
    Exponential request retry strategy.

    :param base: exponentially growing delay base
    :param factor: exponentially growing delay factor (multiplier)
    :param max_value: delay max value
    """

    base: float = 1.0
    factor: float = 2.0
    max_value: Optional[float] = None

    def __call__(self) -> Iterator[float]:
        def gen() -> Generator[float, None, None]:
            for attempt, base in enumerate(it.repeat(self.base, self.attempts)):
                value = base * (self.factor ** attempt) + self.jitter(attempt)
                yield min(self.max_value, value) if self.max_value is not None else value

        return gen()


@dc.dataclass(frozen=True)
class FibonacciBackoff(Backoff):
    """
    Fibonacci request retry strategy.

    :param multiplier: fibonacci interval sequence multiplier
    :param max_value: delay max value
    """

    multiplier: float = 1.0
    max_value: float = 1.0

    def __call__(self) -> Iterator[float]:
        def gen() -> Generator[float, None, None]:
            prev, cur = 1, 1

            for attempt in range(self.attempts):
                value = cur * self.multiplier + self.jitter(attempt)
                yield min(self.max_value, value) if self.max_value is not None else value

                tmp = cur
                cur = prev + cur
                prev = tmp

        return gen()


@dc.dataclass(frozen=True)
class RetryStrategy:
    """
    JSON-RPC request retry strategy.

    :param backoff: backoff delay generator
    :param codes: JSON-RPC response codes receiving which the request will be retried
    :param exceptions: exceptions catching which the request will be retried
    """

    backoff: Backoff
    codes: Optional[set[int]] = None
    exceptions: Optional[set[type[Exception]]] = None


class RetryMiddleware:
    def __init__(self, retry_strategy: RetryStrategy):
        self._retry_strategy = retry_strategy

    def __call__(
        self,
        request: AbstractRequest,
        request_kwargs: Mapping[str, Any],
        /,
        handler: MiddlewareHandler,
    ) -> Optional[AbstractResponse]:
        """
        Request retrying middleware
        """

        delays = self._retry_strategy.backoff()

        for attempt in it.count(start=1):
            try:
                response = handler(request, request_kwargs)
                if response is not None and response.is_error and self._retry_strategy.codes:
                    if (code := response.unwrap_error().code) in self._retry_strategy.codes:
                        delay = next(delays, None)
                        if delay is not None:
                            logger.debug("retrying request: attempt=%d, code=%s", attempt, code)
                            time.sleep(delay)
                            continue

                return response

            except tuple(self._retry_strategy.exceptions or {}) as e:
                delay = next(delays, None)
                if delay is not None:
                    logger.debug("retrying request: attempt=%d, exception=%r", attempt, e)
                    time.sleep(delay)
                else:
                    raise e
        else:
            raise AssertionError("unreachable")


class AsyncRetryMiddleware:
    def __init__(self, retry_strategy: RetryStrategy):
        self._retry_strategy = retry_strategy

    async def __call__(
        self,
        request: AbstractRequest,
        request_kwargs: Mapping[str, Any],
        /,
        handler: AsyncMiddlewareHandler,
    ) -> Optional[AbstractResponse]:
        """
        Asynchronous request retrying middleware
        """

        delays = self._retry_strategy.backoff()

        for attempt in it.count(start=1):
            try:
                response = await handler(request, request_kwargs)
                if response is not None and response.is_error and self._retry_strategy.codes:
                    if (code := response.unwrap_error().code) in self._retry_strategy.codes:
                        delay = next(delays, None)
                        if delay is not None:
                            logger.debug("retrying request: attempt=%d, code=%s", attempt, code)
                            await asyncio.sleep(delay)
                            continue

                return response

            except tuple(self._retry_strategy.exceptions or {}) as e:
                delay = next(delays, None)
                if delay is not None:
                    logger.debug("retrying request: attempt=%d, exception=%r", attempt, e)
                    await asyncio.sleep(delay)
                else:
                    raise e
        else:
            raise AssertionError("unreachable")



"""
OpenRPC specification. See https://spec.open-rpc.org/.
"""

import dataclasses as dc
import enum
from typing import Any, Union

from pjrpc.common import UNSET, MaybeSet

Json = Union[str, int, float, dict[str, 'Json'], bool, list['Json'], tuple['Json'], set['Json'], None]
JsonSchema = dict[str, Json]


@dc.dataclass
class Reference:
    """
    A simple object to allow referencing other components in the specification, internally and externally.
    :param ref: the reference identifier.
    """

    ref: str

    def __post_init__(self) -> None:
        self.__dict__['$ref'] = self.__dict__['ref']
        self.__dataclass_fields__['ref'].name = '$ref'  # noqa


@dc.dataclass
class Contact:
    """
    Contact information for the exposed API.

    :param name: the identifying name of the contact person/organization
    :param url: the URL pointing to the contact information
    :param email: the email address of the contact person/organization
    """

    name: MaybeSet[str] = UNSET
    url: MaybeSet[str] = UNSET
    email: MaybeSet[str] = UNSET


@dc.dataclass
class License:
    """
    License information for the exposed API.

    :param name: the license name used for the API
    :param url: a URL to the license used for the API
    """

    name: str
    url: MaybeSet[str] = UNSET


@dc.dataclass
class Info:
    """
    Metadata about the API.

    :param title: the title of the application
    :param version: the version of the OpenRPC document
    :param description: a verbose description of the application
    :param contact: the contact information for the exposed API
    :param license: the license information for the exposed API
    :param termsOfService: a URL to the Terms of Service for the API
    """

    title: str
    version: str
    description: MaybeSet[str] = UNSET
    contact: MaybeSet[Contact] = UNSET
    license: MaybeSet[License] = UNSET
    termsOfService: MaybeSet[str] = UNSET


@dc.dataclass
class ServerVariable:
    """
    An object representing a Server Variable for server URL template substitution.

    :param default: The default value to use for substitution.
    :param enum: An enumeration of string values to be used if the substitution options are from a limited set.
    :param description: An optional description for the server variable.
    """

    default: str
    enum: MaybeSet[list[str]] = UNSET
    description: MaybeSet[str] = UNSET


@dc.dataclass
class Server:
    """
    Connectivity information of a target server.

    :param name: a name to be used as the canonical name for the server.
    :param url: a URL to the target host. This URL supports Server Variables.
    :param summary: a short summary of what the server is
    :param description: an optional string describing the host designated by the URL
    :param variables: A map between a variable name and its value.
                      The value is passed into the Runtime Expression to produce a server URL.
    """

    name: str
    url: str
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    variables: MaybeSet[dict[str, ServerVariable]] = UNSET


@dc.dataclass
class ExternalDocumentation:
    """
    Allows referencing an external resource for extended documentation.

    :param url: A verbose explanation of the target documentation
    :param description: The URL for the target documentation. Value MUST be in the format of a URL
    """

    url: str
    description: MaybeSet[str] = UNSET


@dc.dataclass
class Tag:
    """
    A list of tags for API documentation control.
    Tags can be used for logical grouping of methods by resources or any other qualifier.

    :param name: the name of the tag
    :param description: a verbose explanation for the tag
    :param externalDocs: additional external documentation for this tag
    """

    name: str
    description: MaybeSet[str] = UNSET
    externalDocs: MaybeSet[ExternalDocumentation] = UNSET


@dc.dataclass
class ExampleObject:
    """
    The ExampleObject object is an object the defines an example.

    :param value: embedded literal example
    :param name: canonical name of the example
    :param summary: short description for the example
    :param description: a verbose explanation of the example
    :param externalValue: a URL that points to the literal example.
    """

    value: Any
    name: str
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    externalValue: MaybeSet[str] = UNSET


@dc.dataclass
class MethodExample:
    """
    The example Pairing object consists of a set of example params and result.

    :param params: example parameters
    :param result: example result
    :param name: name for the example pairing
    :param summary: short description for the example pairing
    :param description: a verbose explanation of the example pairing
    """

    name: str
    params: list[Union[ExampleObject, Reference]]
    result: Union[ExampleObject, Reference]
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET


@dc.dataclass
class ContentDescriptor:
    """
    Content Descriptors are objects that describe content.
    They are reusable ways of describing either parameters or result.

    :param name: name of the content that is being described
    :param schema: schema that describes the content. The Schema Objects MUST follow the specifications outline
                   in the JSON Schema Specification 7 (https://json-schema.org/draft-07/json-schema-release-notes.html)
    :param summary: a short summary of the content that is being described
    :param description: a verbose explanation of the content descriptor behavior
    :param required: determines if the content is a required field
    :param deprecated: specifies that the content is deprecated and SHOULD be transitioned out of usage
    """

    name: str
    schema: JsonSchema
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    required: MaybeSet[bool] = UNSET
    deprecated: MaybeSet[bool] = UNSET


@dc.dataclass
class Error:
    """
    Defines an application level error.

    :param code: a Number that indicates the error type that occurred
    :param message: a String providing a short description of the error
    :param data: a Primitive or Structured value that contains additional information about the error
    """

    code: int
    message: str
    data: MaybeSet[Any] = UNSET


class ParamStructure(str, enum.Enum):
    """
    The expected format of the parameters.
    """

    BY_NAME = 'by-name'
    BY_POSITION = 'by-position'
    EITHER = 'either'


@dc.dataclass
class Link:
    """
    :param name: Canonical name of the link.
    :param description: A description of the link.
    :param summary: Short description for the link.
    :param method: The name of an existing, resolvable OpenRPC method, as defined with a unique method.
    :param params: A map representing parameters to pass to a method as specified with method.
                   The key is the parameter name to be used, whereas the value can be a constant or a runtime
                   expression to be evaluated and passed to the linked method.
    :param server: A server object to be used by the target method.
    """

    name: str
    description: MaybeSet[str] = UNSET
    summary: MaybeSet[str] = UNSET
    method: MaybeSet[str] = UNSET
    params: MaybeSet[dict[str, Any]] = UNSET
    server: MaybeSet[Server] = UNSET


@dc.dataclass
class MethodInfo:
    """
    Describes the interface for the given method name.

    :param name: the canonical name for the method
    :param params: a list of parameters that are applicable for this method
    :param result: the description of the result returned by the method
    :param errors: a list of custom application defined errors that MAY be returned
    :param examples: method usage examples
    :param summary: a short summary of what the method does
    :param description: a verbose explanation of the method behavior
    :param tags: a list of tags for API documentation control
    :param deprecated: declares this method to be deprecated
    :param paramStructure: the expected format of the parameters
    :param externalDocs: additional external documentation for this method
    :param servers: an alternative servers array to service this method
    """

    name: str
    params: list[Union[ContentDescriptor, Reference]]
    result: Union[ContentDescriptor, Reference]
    errors: MaybeSet[list[Union[Error, Reference]]] = UNSET
    links: MaybeSet[list[Union[Link, Reference]]] = UNSET
    paramStructure: MaybeSet[ParamStructure] = UNSET
    examples: MaybeSet[list[Union[MethodExample, Reference]]] = UNSET
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    tags: MaybeSet[list[Union[Tag, Reference]]] = UNSET
    deprecated: MaybeSet[bool] = UNSET
    externalDocs: MaybeSet[ExternalDocumentation] = UNSET
    servers: MaybeSet[list[Server]] = UNSET


@dc.dataclass
class Components:
    """
    Set of reusable objects for different aspects of the OpenRPC.

    :param contentDescriptors: reusable Schema Objects
    :param schemas: reusable Schema Objects
    :param examples: reusable Schema Objects
    :param links: reusable Schema Objects
    :param errors: reusable Schema Objects
    :param examplePairingObjects: reusable Schema Objects
    :param tags: reusable Schema Objects
    """

    contentDescriptors: MaybeSet[dict[str, ContentDescriptor]] = UNSET
    schemas: MaybeSet[dict[str, JsonSchema]] = UNSET
    examples: MaybeSet[dict[str, ExampleObject]] = UNSET
    links: MaybeSet[dict[str, Link]] = UNSET
    errors: MaybeSet[dict[str, Error]] = UNSET
    examplePairingObjects: MaybeSet[dict[str, MethodExample]] = UNSET
    tags: MaybeSet[dict[str, Tag]] = UNSET


@dc.dataclass
class SpecRoot:
    """
    The root object of the OpenRPC document.

    :param info: provides metadata about the API
    :param components: an element to hold various schemas for the specification.
    :param methods: the available methods for the API
    :param servers: connectivity information
    :param externalDocs: additional external documentation
    :param openrpc: the semantic version number of the OpenRPC Specification version that the OpenRPC document uses
    """

    info: Info
    components: Components
    methods: list[Union[MethodInfo, Reference]] = dc.field(default_factory=list)
    servers: MaybeSet[list[Server]] = UNSET
    externalDocs: MaybeSet[ExternalDocumentation] = UNSET
    openrpc: str = '1.3.2'



"""
JSON-RPC method parameters validators.
"""

from .base import BaseValidator, BaseValidatorFactory, ExcludeFunc, ValidationError

__all__ = [
    'BaseValidator',
    'BaseValidatorFactory',
    'ExcludeFunc',
    'ValidationError',
]



import enum
from typing import Any, Literal, TypeVar, Union


class UnsetType(enum.Enum):
    """
    `Sentinel <https://en.wikipedia.org/wiki/Sentinel_value>`_ object.
    Used to distinct unset (missing) values from ``None`` ones.
    """

    UNSET = "UNSET"

    def __bool__(self) -> Literal[False]:
        return False

    def __repr__(self) -> str:
        return "UNSET"

    def __str__(self) -> str:
        return repr(self)


UNSET: UnsetType = UnsetType.UNSET

MaybeSetType = TypeVar('MaybeSetType')
MaybeSet = Union[UnsetType, MaybeSetType]

JsonT = Any



import json
import logging
from typing import Any, Callable, Iterable, Optional

import aio_pika
from yarl import URL

import pjrpc.server
from pjrpc.server.dispatcher import AsyncExecutor, AsyncMiddlewareType, JSONEncoder

logger = logging.getLogger(__package__)

AioPikaDispatcher = pjrpc.server.AsyncDispatcher[aio_pika.abc.AbstractIncomingMessage]


class Executor:
    """
    `aio_pika <https://aio-pika.readthedocs.io/en/latest/>`_ based JSON-RPC server.

    :param broker_url: broker connection url
    :param request_queue_name: requests queue name
    :param response_exchange_name: response exchange name
    :param response_routing_key: response routing key
    :param prefetch_count: worker prefetch count
    """

    def __init__(
        self,
        broker_url: URL,
        request_queue_name: str,
        request_queue_args: Optional[dict[str, Any]] = None,
        response_exchange_name: Optional[str] = None,
        response_exchange_args: Optional[dict[str, Any]] = None,
        response_routing_key: Optional[str] = None,
        prefetch_count: int = 0,
        executor: Optional[AsyncExecutor] = None,
        json_loader: Callable[..., Any] = json.loads,
        json_dumper: Callable[..., str] = json.dumps,
        json_encoder: type[JSONEncoder] = JSONEncoder,
        json_decoder: Optional[type[json.JSONDecoder]] = None,
        middlewares: Iterable[AsyncMiddlewareType[aio_pika.abc.AbstractIncomingMessage]] = (),
        max_batch_size: Optional[int] = None,
    ):
        self._broker_url = broker_url
        self._request_queue_name = request_queue_name
        self._request_queue_args = request_queue_args
        self._response_exchange_name = response_exchange_name
        self._response_exchange_args = response_exchange_args
        self._response_routing_key = response_routing_key
        self._prefetch_count = prefetch_count

        self._connection = aio_pika.connection.Connection(broker_url)
        self._channel: Optional[aio_pika.abc.AbstractChannel] = None

        self._request_queue: Optional[aio_pika.abc.AbstractQueue] = None
        self._response_exchange: Optional[aio_pika.abc.AbstractExchange] = None
        self._consumer_tag: Optional[str] = None
        self._dispatcher = AioPikaDispatcher(
            executor=executor,
            json_loader=json_loader,
            json_dumper=json_dumper,
            json_encoder=json_encoder,
            json_decoder=json_decoder,
            middlewares=middlewares,
            max_batch_size=max_batch_size,
        )

    @property
    def dispatcher(self) -> AioPikaDispatcher:
        """
        JSON-RPC method dispatcher.
        """

        return self._dispatcher

    async def start(self) -> None:
        """
        Starts executor.
        """

        await self._connection.connect()
        self._channel = channel = await self._connection.channel()
        await channel.set_qos(prefetch_count=self._prefetch_count)

        self._request_queue = await channel.declare_queue(
            self._request_queue_name, **(self._request_queue_args or {}),
        )

        if self._response_exchange_name:
            self._response_exchange = await channel.declare_exchange(
                self._response_exchange_name, **(self._response_exchange_args or {}),
            )

        self._consumer_tag = await self._request_queue.consume(self._rpc_handle)

    async def shutdown(self) -> None:
        """
        Stops executor.
        """

        assert self._channel and self._request_queue and self._consumer_tag, "executor not started"

        await self._request_queue.cancel(self._consumer_tag)
        await self._channel.close()
        await self._connection.close()

    async def _rpc_handle(self, message: aio_pika.abc.AbstractIncomingMessage) -> None:
        """
        Handles JSON-RPC request.

        :param message: incoming message
        """

        async with message.process():
            try:
                if (response := await self._dispatcher.dispatch(message.body.decode(), context=message)) is not None:
                    response_text, error_codes = response

                    async with self._connection.channel() as channel:
                        exchange = self._response_exchange or channel.default_exchange
                        await exchange.publish(
                            aio_pika.Message(
                                body=response_text.encode(),
                                correlation_id=message.correlation_id,
                                content_type=pjrpc.common.DEFAULT_CONTENT_TYPE,
                            ),
                            routing_key=message.reply_to or self._response_routing_key or '',
                        )

            except Exception as e:
                logger.exception("jsonrpc request handling error: %s", e)
                raise



import json
import typing
from typing import Any, Callable, Generator, Iterable, Mapping, Optional, Sequence, TypedDict, Union

import httpx

import pjrpc
from pjrpc.client import AbstractAsyncClient, AbstractClient, AsyncMiddleware, Middleware
from pjrpc.common import AbstractRequest, AbstractResponse, BatchRequest, BatchResponse, JSONEncoder, JsonRpcError
from pjrpc.common import Request, Response, generators
from pjrpc.common.typedefs import JsonRpcRequestIdT


class RequestArgs(TypedDict, total=False):
    headers: Union[Mapping[str, str], Sequence[tuple[str, str]]]
    cookies: httpx._types.CookieTypes
    auth: Union[httpx._types.AuthTypes, httpx._client.UseClientDefault]
    follow_redirects: Union[bool, httpx._client.UseClientDefault]
    timeout: Union[httpx._types.TimeoutTypes, httpx._client.UseClientDefault]
    extensions: httpx._types.RequestExtensions


class Client(AbstractClient):
    """
    `httpx <https://www.python-httpx.org/>`_ library sync client backend.

    :param url: url to be used as JSON-RPC endpoint.
    :param http_client: custom client to be used instead of :py:class:`httpx.Client`
    :param raise_for_status: should `ClientResponse.raise_for_status()` be called automatically
    :param id_gen_impl: identifier generator
    :param error_cls: JSON-RPC error base class
    :param json_loader: json loader
    :param json_dumper: json dumper
    :param json_encoder: json encoder
    :param json_decoder: json decoder
    """

    def __init__(
        self,
        url: str,
        *,
        http_client: Optional[httpx.Client] = None,
        raise_for_status: bool = True,
        id_gen_impl: Callable[..., Generator[JsonRpcRequestIdT, None, None]] = generators.sequential,
        error_cls: type[JsonRpcError] = JsonRpcError,
        json_loader: Callable[..., Any] = json.loads,
        json_dumper: Callable[..., str] = json.dumps,
        json_encoder: type[JSONEncoder] = JSONEncoder,
        json_decoder: Optional[json.JSONDecoder] = None,
        middlewares: Iterable[Middleware] = (),
    ):
        super().__init__(
            id_gen_impl=id_gen_impl,
            error_cls=error_cls,
            json_loader=json_loader,
            json_dumper=json_dumper,
            json_encoder=json_encoder,
            json_decoder=json_decoder,
            middlewares=middlewares,
        )
        self._endpoint = url
        self._http_client = http_client or httpx.Client()
        self._owned_http_client = http_client is None
        self._raise_for_status = raise_for_status

    @typing.overload
    def send(self, request: Request, **kwargs: Any) -> Optional[Response]:
        ...

    @typing.overload
    def send(self, request: BatchRequest, **kwargs: Any) -> Optional[BatchResponse]:
        ...

    def send(self, request: AbstractRequest, **kwargs: Any) -> Optional[AbstractResponse]:
        """
        Sends a JSON-RPC request.

        :param request: request instance
        :param kwargs: additional client request argument
        :returns: response instance or None if the request is a notification
        """

        return self._send(request, kwargs)

    def _request(
        self,
        request_text: str,
        is_notification: bool,
        request_kwargs: Mapping[str, Any],
    ) -> Optional[str]:
        """
        Sends a JSON-RPC request.

        :param request_text: request text
        :param is_notification: is the request a notification
        :param request_kwargs: additional client request argument
        :returns: response text
        """

        request_kwargs = typing.cast(RequestArgs, request_kwargs)

        request_kwargs['headers'] = headers = dict(request_kwargs.get('headers', {}))
        headers['Content-Type'] = self._request_content_type

        resp = self._http_client.post(self._endpoint, content=request_text, **request_kwargs)
        if self._raise_for_status:
            resp.raise_for_status()
        if is_notification:
            return None

        response_text = resp.text
        content_type = resp.headers.get('Content-Type', '')
        if response_text and content_type.split(';')[0] not in self._response_content_types:
            raise pjrpc.exc.DeserializationError(f"unexpected response content type: {content_type}")

        return response_text

    def close(self) -> None:
        """
        Closes the current http session.
        """

        if self._owned_http_client:
            self._http_client.close()

    def __enter__(self) -> 'Client':
        if self._owned_http_client:
            self._http_client.__enter__()

        return self

    def __exit__(self, *args: Any) -> None:
        if self._owned_http_client:
            self._http_client.__exit__(*args)


class AsyncClient(AbstractAsyncClient):
    """
    `httpx <https://www.python-httpx.org/>`_ library async client backend.

    :param url: url to be used as JSON-RPC endpoint.
    :param http_client: custom client to be used instead of :py:class:`httpx.AsyncClient`
    :param raise_for_status: should `ClientResponse.raise_for_status()` be called automatically
    :param id_gen_impl: identifier generator
    :param error_cls: JSON-RPC error base class
    :param json_loader: json loader
    :param json_dumper: json dumper
    :param json_encoder: json encoder
    :param json_decoder: json decoder
    """

    def __init__(
        self,
        url: str,
        *,
        http_client: Optional[httpx.AsyncClient] = None,
        raise_for_status: bool = True,
        id_gen_impl: Callable[..., Generator[JsonRpcRequestIdT, None, None]] = generators.sequential,
        error_cls: type[JsonRpcError] = JsonRpcError,
        json_loader: Callable[..., Any] = json.loads,
        json_dumper: Callable[..., str] = json.dumps,
        json_encoder: type[JSONEncoder] = JSONEncoder,
        json_decoder: Optional[json.JSONDecoder] = None,
        middlewares: Iterable[AsyncMiddleware] = (),
    ):
        super().__init__(
            id_gen_impl=id_gen_impl,
            error_cls=error_cls,
            json_loader=json_loader,
            json_dumper=json_dumper,
            json_encoder=json_encoder,
            json_decoder=json_decoder,
            middlewares=middlewares,
        )
        self._endpoint = url
        self._http_client = http_client or httpx.AsyncClient()
        self._owned_http_client = http_client is None
        self._raise_for_status = raise_for_status

    @typing.overload
    async def send(self, request: Request, **kwargs: Any) -> Optional[Response]:
        ...

    @typing.overload
    async def send(self, request: BatchRequest, **kwargs: Any) -> Optional[BatchResponse]:
        ...

    async def send(self, request: AbstractRequest, **kwargs: Any) -> Optional[AbstractResponse]:
        """
        Sends a JSON-RPC request.

        :param request: request instance
        :param kwargs: additional client request argument
        :returns: response instance or None if the request is a notification
        """

        return await self._send(request, kwargs)

    async def _request(
        self,
        request_text: str,
        is_notification: bool,
        request_kwargs: Mapping[str, Any],
    ) -> Optional[str]:
        """
        Sends a JSON-RPC request.

        :param request_text: request text
        :param is_notification: is the request a notification
        :param request_kwargs: additional client request argument
        :returns: response text
        """

        request_kwargs = typing.cast(RequestArgs, request_kwargs)

        request_kwargs['headers'] = headers = dict(request_kwargs.get('headers', {}))
        headers['Content-Type'] = self._request_content_type

        resp = await self._http_client.post(self._endpoint, content=request_text, **request_kwargs)
        if self._raise_for_status:
            resp.raise_for_status()

        response_buff: list[str] = []
        async for chunk in resp.aiter_text():
            response_buff.append(chunk)

        response_text = ''.join(response_buff)
        if is_notification:
            return None

        content_type = resp.headers.get('Content-Type', '')
        if response_text and content_type.split(';')[0] not in self._response_content_types:
            raise pjrpc.exc.DeserializationError(f"unexpected response content type: {content_type}")

        return response_text

    async def close(self) -> None:
        """
        Closes current http session.
        """

        if self._owned_http_client:
            await self._http_client.aclose()

    async def __aenter__(self) -> 'AsyncClient':
        if self._owned_http_client:
            await self._http_client.__aenter__()

        return self

    async def __aexit__(self, *args: Any) -> None:
        if self._owned_http_client:
            await self._http_client.__aexit__(*args)



"""
Flask JSON-RPC extension.
"""

import functools as ft
import json
from typing import Any, Callable, Iterable, Optional, Union

import flask
from flask import current_app
from werkzeug import exceptions

import pjrpc.server
from pjrpc.server import specs, utils
from pjrpc.server.dispatcher import Executor, JSONEncoder, MethodRegistry, MiddlewareType

FlaskDispatcher = pjrpc.server.Dispatcher[None]


class JsonRPC:
    """
    `Flask <https://flask.palletsprojects.com/en/1.1.x/>`_ framework JSON-RPC extension class.

    :param prefix: JSON-RPC handler base path
    :param status_by_error: a function returns http status code by json-rpc error codes, 200 for all errors by default
    """

    def __init__(
        self,
        prefix: str = '',
        http_app: Optional[Union[flask.Flask, flask.Blueprint]] = None,
        status_by_error: Callable[[tuple[int, ...]], int] = lambda codes: 200,
        executor: Optional[Executor] = None,
        json_loader: Callable[..., Any] = json.loads,
        json_dumper: Callable[..., str] = json.dumps,
        json_encoder: type[JSONEncoder] = JSONEncoder,
        json_decoder: Optional[type[json.JSONDecoder]] = None,
        middlewares: Iterable[MiddlewareType[None]] = (),
        max_batch_size: Optional[int] = None,
    ):
        self._prefix = prefix.rstrip('/')
        self._http_app = http_app or flask.Flask(__name__)
        self._status_by_error = status_by_error

        self._executor = executor
        self._json_loader = json_loader
        self._json_dumper = json_dumper
        self._json_encoder = json_encoder
        self._json_decoder = json_decoder
        self._middlewares = middlewares
        self._max_batch_size = max_batch_size

        self._endpoints: dict[str, FlaskDispatcher] = {}
        self._subapps: dict[str, JsonRPC] = {}

    @property
    def http_app(self) -> Union[flask.Flask, flask.Blueprint]:
        """
        aiohttp application.
        """

        return self._http_app

    @property
    def endpoints(self) -> dict[str, FlaskDispatcher]:
        """
        JSON-RPC application registered endpoints.
        """

        return self._endpoints

    def add_methods(self, registry: MethodRegistry, endpoint: str = '') -> 'JsonRPC':
        """
        Adds methods to the provided endpoint.

        :param registry: methods registry
        :param endpoint: endpoint path
        """

        dispatcher = self._get_endpoint(endpoint)
        dispatcher.add_methods(registry)
        return self

    def add_subapp(self, prefix: str, subapp: 'JsonRPC') -> None:
        """
        Adds sub-application accessible under provided prefix.

        :param prefix: path under which sub-application is accessed.
        :param subapp: sub-application instance
        """

        assert isinstance(subapp.http_app, flask.Blueprint), "subapp must be a sub-application instance"

        prefix = prefix.rstrip('/')
        if not prefix:
            raise ValueError("prefix cannot be empty")

        for dispatcher in subapp.endpoints.values():
            dispatcher.add_middlewares(*self._middlewares, before=True)

        self._http_app.register_blueprint(subapp.http_app, url_prefix=utils.join_path(self._prefix, prefix))
        self._subapps[prefix] = subapp

    def add_spec(self, spec: specs.Specification, endpoint: str = '', path: str = '') -> None:
        """
        Adds JSON-RPC specification of the provided endpoint to the provided path.

        :param spec: JSON-RPC specification
        :param endpoint: specification endpoint
        :param path: path under witch the specification will be accessible.
        """

        self._http_app.add_url_rule(
            f"/{utils.join_path(self._prefix, endpoint, path)}",
            methods=['GET'],
            endpoint=self._get_spec.__name__,
            view_func=ft.partial(self._get_spec, endpoint=endpoint, spec=spec, path=path),
        )

    def add_spec_ui(self, path: str, ui: specs.BaseUI, spec_url: str) -> None:
        """
        Adds JSON-RPC specification ui.

        :param path: path under which ui will be accessible.
        :param ui: specification ui instance
        :param spec_url: specification url
        """

        ui_app = flask.Blueprint(ui.__class__.__name__, __name__)
        ui_app.add_url_rule(
            '/',
            methods=['GET'],
            endpoint=self._ui_index_page.__name__,
            view_func=ft.partial(self._ui_index_page, ui=ui, spec_url=spec_url),
        )
        ui_app.add_url_rule(
            '/index.html',
            methods=['GET'],
            endpoint=f'{self._ui_index_page.__name__}-index',
            view_func=ft.partial(self._ui_index_page, ui=ui, spec_url=spec_url),
        )
        ui_app.add_url_rule(
            '/<path:filename>',
            methods=['GET'],
            endpoint=self._ui_static.__name__,
            view_func=ft.partial(self._ui_static, ui=ui),
        )

        self._http_app.register_blueprint(ui_app, url_prefix=utils.join_path(self._prefix, path))

    def generate_spec(self, spec: specs.Specification, base_path: str = '', endpoint: str = '') -> dict[str, Any]:
        """
        Generates JSON-RPC specification of the provided endpoint.

        :param spec: JSON-RPC specification
        :param base_path: specification base path
        :param endpoint: endpoint the specification is generated for
        """

        app_endpoints = self._endpoints
        for prefix, subapp in self._subapps.items():
            for subprefix, dispatcher in subapp.endpoints.items():
                app_endpoints[utils.join_path(prefix, subprefix)] = dispatcher

        methods = {
            utils.remove_prefix(dispatcher_endpoint, endpoint): dispatcher.registry.values()
            for dispatcher_endpoint, dispatcher in app_endpoints.items()
            if dispatcher_endpoint.startswith(endpoint)
        }
        return spec.generate(
            root_endpoint=utils.join_path(base_path, endpoint),
            methods=methods,
        )

    def _ui_index_page(self, ui: specs.BaseUI, spec_url: str) -> flask.Response:
        return current_app.response_class(response=ui.get_index_page(spec_url), content_type='text/html')

    def _ui_static(self, ui: specs.BaseUI, filename: str) -> flask.Response:
        return flask.send_from_directory(ui.get_static_folder(), filename)

    def _get_endpoint(self, endpoint: str) -> FlaskDispatcher:
        endpoint = endpoint.rstrip('/')

        if endpoint not in self._endpoints:
            self._endpoints[endpoint] = dispatcher = FlaskDispatcher(
                executor=self._executor,
                json_loader=self._json_loader,
                json_dumper=self._json_dumper,
                json_encoder=self._json_encoder,
                json_decoder=self._json_decoder,
                middlewares=self._middlewares,
                max_batch_size=self._max_batch_size,
            )
            self._http_app.add_url_rule(
                f"/{utils.join_path(self._prefix, endpoint)}",
                methods=['POST'],
                endpoint="rpc_handle",
                view_func=ft.partial(self._rpc_handle, dispatcher=dispatcher),
            )
        else:
            dispatcher = self._endpoints[endpoint]

        return dispatcher

    def _get_spec(self, endpoint: str, spec: specs.Specification, path: str) -> flask.Response:
        base_path = utils.remove_suffix(flask.request.path, suffix=utils.join_path(endpoint, path))
        schema = self.generate_spec(base_path=base_path, endpoint=endpoint.rstrip('/'), spec=spec)

        return current_app.response_class(
            self._json_dumper(schema, cls=self._json_encoder),
            mimetype=pjrpc.common.DEFAULT_CONTENT_TYPE,
        )

    def _rpc_handle(self, dispatcher: FlaskDispatcher) -> flask.Response:
        """
        Handles JSON-RPC request.

        :returns: flask response
        """

        if not flask.request.is_json:
            raise exceptions.UnsupportedMediaType()

        try:
            request_text = flask.request.get_data(as_text=True)
        except UnicodeDecodeError as e:
            raise exceptions.BadRequest() from e

        response = dispatcher.dispatch(request_text, context=None)
        if response is None:
            return current_app.response_class()
        else:
            response_text, error_codes = response
            return current_app.response_class(
                response_text,
                status=self._status_by_error(error_codes),
                mimetype=pjrpc.common.DEFAULT_CONTENT_TYPE,
            )



import functools as ft
import pathlib
import re
from typing import Any

import openapi_ui_bundles

from pjrpc.server.specs import BaseUI


class SwaggerUI(BaseUI):
    """
    Swagger UI.

    :param config: documentation configurations
                   (see https://github.com/swagger-api/swagger-ui/blob/master/docs/usage/configuration.md).
    """

    def __init__(self, **configs: Any):
        self._configs = configs
        self._static_folder = pathlib.Path(openapi_ui_bundles.swagger_ui.static_path)

    def get_static_folder(self) -> pathlib.Path:
        return self._static_folder

    @ft.lru_cache
    def get_index_page(self, spec_url: str) -> str:
        index_path = self.get_static_folder() / 'index.html'
        index_page = index_path.read_text()

        config = dict(self._configs, **{'url': spec_url, 'dom_id': '#swagger-ui'})
        config_str = ', '.join(f'{param}: "{value}"' for param, value in config.items())

        return re.sub(
            pattern=r'SwaggerUIBundle\({.*?}\)',
            repl=f'SwaggerUIBundle({{ {config_str} }})',
            string=index_page,
            count=1,
            flags=re.DOTALL,
        )


class RapiDoc(BaseUI):
    """
    RapiDoc UI.

    :param config: documentation configurations (see https://mrin9.github.io/RapiDoc/api.html).
                   Be aware that configuration parameters should be in snake case,
                   for example: parameter `heading-text` should be passed as `heading_text`)
    """

    def __init__(self, **configs: Any):
        self._configs = configs
        self._static_folder = pathlib.Path(openapi_ui_bundles.rapidoc.static_path)

    def get_static_folder(self) -> pathlib.Path:
        return self._static_folder

    @ft.lru_cache
    def get_index_page(self, spec_url: str) -> str:
        index_path = pathlib.Path(self.get_static_folder()) / 'index.html'
        index_page = index_path.read_text()

        config = dict(self._configs, **{'spec_url': spec_url, 'id': 'thedoc'})
        config_str = ' '.join(f'{param.replace("_", "-")}="{value}"' for param, value in config.items())

        return re.sub(
            pattern='<rapi-doc.*?>',
            repl=f'<rapi-doc {config_str}>',
            string=index_page,
            count=1,
            flags=re.DOTALL,
        )


class ReDoc(BaseUI):
    """
    ReDoc UI.

    :param config: documentation configurations (see https://github.com/Redocly/redoc#configuration).
                   Be aware that configuration parameters should be in snake case,
                   for example: parameter `heading-text` should be passed as `heading_text`)
    """

    def __init__(self, **configs: Any):
        self._configs = configs
        self._static_folder = pathlib.Path(openapi_ui_bundles.redoc.static_path)

    def get_static_folder(self) -> pathlib.Path:
        return self._static_folder

    @ft.lru_cache
    def get_index_page(self, spec_url: str) -> str:
        index_path = pathlib.Path(self.get_static_folder()) / 'index.html'
        index_page = index_path.read_text()

        config = dict(self._configs, **{'spec_url': spec_url})
        config_str = ' '.join(f'{param.replace("_", "-")}="{value}"' for param, value in config.items())

        return re.sub(
            pattern='<redoc.*?>',
            repl=f'<redoc {config_str}>',
            string=index_page,
            count=1,
            flags=re.DOTALL,
        )



import functools as ft
from typing import Generator

import pytest

from .pytest import PjRpcMocker

PjRpcRequestsMocker = ft.partial(PjRpcMocker, target='pjrpc.client.backend.requests.Client._request')


@pytest.fixture
def pjrpc_requests_mocker() -> Generator[PjRpcMocker, None, None]:
    """
    Requests client mocking fixture.
    """

    with PjRpcRequestsMocker() as mocker:
        yield mocker



from typing import Any, Callable, Optional


def remove_prefix(s: str, prefix: str) -> str:
    """
    Removes a prefix from a string.

    :param s: string to be processed
    :param prefix: prefix to be removed
    :return: processed string
    """

    if s.startswith(prefix):
        return s[len(prefix):]
    else:
        return s


def remove_suffix(s: str, suffix: str) -> str:
    """
    Removes a suffix from a string.

    :param s: string to be processed
    :param suffix: suffix to be removed
    :return: processed string
    """

    if suffix and s.endswith(suffix):
        return s[0:-len(suffix)]
    else:
        return s


def join_path(path: str, *paths: str) -> str:
    result = path
    for path in paths:
        if path:
            result = f'{result.rstrip("/")}/{path.lstrip("/")}'

    return result


ExcludeFunc = Callable[[int, str, Optional[type[Any]], Optional[Any]], bool]


def exclude_positional_param(param_index: int) -> ExcludeFunc:
    def exclude(index: int, name: str, typ: Optional[type[Any]], default: Optional[Any]) -> bool:
        return index == param_index

    return exclude


def exclude_named_param(param_name: str) -> ExcludeFunc:
    def exclude(index: int, name: str, typ: Optional[type[Any]], default: Optional[Any]) -> bool:
        return name == param_name

    return exclude



"""
aiohttp JSON-RPC server integration.
"""

import functools as ft
import inspect
import json
from typing import Any, Callable, Iterable, Mapping, Optional

from aiohttp import web

import pjrpc
from pjrpc.server import specs, utils
from pjrpc.server.dispatcher import AsyncExecutor, AsyncMiddlewareType, JSONEncoder, MethodRegistry


def is_aiohttp_request(idx: int, name: str, annotation: Optional[type[Any]], default: Optional[Any]) -> bool:
    if annotation is None:
        return False

    return inspect.isclass(annotation) and issubclass(annotation, web.Request)


AioHttpDispatcher = pjrpc.server.AsyncDispatcher[web.Request]


class Application:
    """
    `aiohttp <https://aiohttp.readthedocs.io/en/stable/web.html>`_ based JSON-RPC server.

    :param prefix: JSON-RPC handler base path
    :param http_app: aiohttp application instance
    :param status_by_error: a function returns http status code by json-rpc error codes, 200 for all errors by default
    """

    def __init__(
        self,
        prefix: str = '',
        http_app: Optional[web.Application] = None,
        status_by_error: Callable[[tuple[int, ...]], int] = lambda codes: 200,
        executor: Optional[AsyncExecutor] = None,
        json_loader: Callable[..., Any] = json.loads,
        json_dumper: Callable[..., str] = json.dumps,
        json_encoder: type[JSONEncoder] = JSONEncoder,
        json_decoder: Optional[type[json.JSONDecoder]] = None,
        middlewares: Iterable[AsyncMiddlewareType[web.Request]] = (),
        max_batch_size: Optional[int] = None,
    ):
        self._prefix = prefix.rstrip('/')
        self._http_app = http_app or web.Application()
        self._status_by_error = status_by_error

        self._executor: Optional[AsyncExecutor] = executor
        self._json_loader: Callable[..., Any] = json_loader
        self._json_dumper: Callable[..., str] = json_dumper
        self._json_encoder: type[JSONEncoder] = json_encoder
        self._json_decoder: Optional[type[json.JSONDecoder]] = json_decoder
        self._middlewares: Iterable[AsyncMiddlewareType[web.Request]] = middlewares
        self._max_batch_size: Optional[int] = max_batch_size

        self._endpoints: dict[str, AioHttpDispatcher] = {}
        self._subapps: dict[str, Application] = {}

    @property
    def http_app(self) -> web.Application:
        """
        aiohttp application.
        """

        return self._http_app

    @property
    def endpoints(self) -> Mapping[str, AioHttpDispatcher]:
        """
        JSON-RPC application registered endpoints.
        """

        return self._endpoints

    def add_methods(self, registry: MethodRegistry, endpoint: str = '') -> 'Application':
        """
        Adds methods to the provided endpoint.

        :param registry: methods registry
        :param endpoint: endpoint path
        """

        dispatcher = self._get_endpoint(endpoint)
        dispatcher.add_methods(registry)
        return self

    def add_subapp(self, prefix: str, subapp: 'Application') -> None:
        """
        Adds sub-application accessible under provided prefix.

        :param prefix: path under which sub-application is accessed.
        :param subapp: sub-application instance
        """

        prefix = prefix.rstrip('/')
        if not prefix:
            raise ValueError("prefix cannot be empty")

        for dispatcher in subapp.endpoints.values():
            dispatcher.add_middlewares(*self._middlewares, before=True)

        self._http_app.add_subapp(utils.join_path(self._prefix, prefix), subapp.http_app)
        self._subapps[prefix] = subapp

    def add_spec(self, spec: specs.Specification, endpoint: str = '', path: str = '') -> None:
        """
        Adds JSON-RPC specification of the provided endpoint to the provided path.

        :param spec: JSON-RPC specification
        :param endpoint: specification endpoint
        :param path: path under witch the specification will be accessible.
        """

        self._http_app.router.add_get(
            utils.join_path(self._prefix, endpoint, path),
            ft.partial(self._get_spec, endpoint=endpoint, spec=spec, path=path),
        )

    def add_spec_ui(self, path: str, ui: specs.BaseUI, spec_url: str) -> None:
        """
        Adds JSON-RPC specification ui.

        :param path: path under which ui will be accessible.
        :param ui: specification ui instance
        :param spec_url: specification url
        """

        ui_app = web.Application()
        ui_app.router.add_get('/', ft.partial(self._ui_index_page, ui=ui, spec_url=spec_url))
        ui_app.router.add_get('/index.html', ft.partial(self._ui_index_page, ui=ui, spec_url=spec_url))
        ui_app.router.add_static('/', ui.get_static_folder())

        self._http_app.add_subapp(utils.join_path(self._prefix, path), ui_app)

    def generate_spec(self, spec: specs.Specification, base_path: str = '', endpoint: str = '') -> dict[str, Any]:
        """
        Generates JSON-RPC specification of the provided endpoint.

        :param spec: JSON-RPC specification
        :param base_path: specification base path
        :param endpoint: endpoint the specification is generated for
        """

        app_endpoints = self._endpoints
        for prefix, subapp in self._subapps.items():
            for subprefix, dispatcher in subapp.endpoints.items():
                app_endpoints[utils.join_path(prefix, subprefix)] = dispatcher

        methods = {
            utils.remove_prefix(dispatcher_endpoint, endpoint): dispatcher.registry.values()
            for dispatcher_endpoint, dispatcher in app_endpoints.items()
            if dispatcher_endpoint.startswith(endpoint)
        }
        return spec.generate(
            root_endpoint=utils.join_path(base_path, endpoint),
            methods=methods,
        )

    async def _ui_index_page(self, request: web.Request, ui: specs.BaseUI, spec_url: str) -> web.Response:
        return web.Response(text=ui.get_index_page(spec_url), content_type='text/html')

    def _get_endpoint(self, endpoint: str) -> AioHttpDispatcher:
        endpoint = endpoint.rstrip('/')

        if endpoint not in self._endpoints:
            self._endpoints[endpoint] = dispatcher = AioHttpDispatcher(
                executor=self._executor,
                json_loader=self._json_loader,
                json_dumper=self._json_dumper,
                json_encoder=self._json_encoder,
                json_decoder=self._json_decoder,
                middlewares=self._middlewares,
                max_batch_size=self._max_batch_size,
            )
            self._http_app.router.add_post(
                utils.join_path(self._prefix, endpoint),
                ft.partial(self._rpc_handle, dispatcher=dispatcher),
            )
        else:
            dispatcher = self._endpoints[endpoint]

        return dispatcher

    async def _get_spec(
        self,
        request: web.Request,
        endpoint: str,
        spec: specs.Specification,
        path: str,
    ) -> web.Response:
        base_path = utils.remove_suffix(request.path, suffix=utils.join_path(endpoint, path))
        schema = self.generate_spec(base_path=base_path, endpoint=endpoint.rstrip('/'), spec=spec)

        return web.json_response(text=self._json_dumper(schema, cls=self._json_encoder))

    async def _rpc_handle(self, http_request: web.Request, dispatcher: AioHttpDispatcher) -> web.Response:
        """
        Handles JSON-RPC request.

        :param http_request: :py:class:`aiohttp.web.Response`
        :returns: :py:class:`aiohttp.web.Request`
        """

        if http_request.content_type not in pjrpc.common.REQUEST_CONTENT_TYPES:
            raise web.HTTPUnsupportedMediaType()

        try:
            request_text = await http_request.text()
        except UnicodeDecodeError as e:
            raise web.HTTPBadRequest() from e

        response = await dispatcher.dispatch(request_text, context=http_request)
        if response is None:
            return web.Response()
        else:
            response_text, error_codes = response
            return web.json_response(status=self._status_by_error(error_codes), text=response_text)



from typing import Awaitable, Callable, Protocol, TypeVar

from pjrpc.common import MaybeSet, Request, Response

__all__ = [
    'AsyncHandlerType',
    'AsyncMiddlewareType',
    'ContextType',
    'HandlerType',
    'MiddlewareResponse',
    'MiddlewareType',
]


ContextType = TypeVar('ContextType')
'''Context argument for RPC methods and middlewares'''  # for sphinx autodoc

AsyncHandlerType = Callable[[Request, ContextType], Awaitable[MaybeSet[Response]]]
'''Async RPC handler method, passed to middlewares'''  # for sphinx autodoc

HandlerType = Callable[[Request, ContextType], MaybeSet[Response]]
'''Blocking RPC handler method, passed to middlewares'''  # for sphinx autodoc

MiddlewareResponse = MaybeSet[Response]
'''middlewares and handlers return Response or UnsetType'''  # for sphinx autodoc


class AsyncMiddlewareType(Protocol[ContextType]):
    """
    Asynchronous middleware type
    """

    async def __call__(
        self,
        request: Request,
        context: ContextType,
        handler: AsyncHandlerType[ContextType],
    ) -> MaybeSet[MiddlewareResponse]:
        pass


class MiddlewareType(Protocol[ContextType]):
    """
    Synchronous middleware type
    """

    async def __call__(
        self,
        request: Request,
        context: ContextType,
        handler: HandlerType[ContextType],
    ) -> MaybeSet[MiddlewareResponse]:
        pass



"""
JSON-RPC client.
"""

from . import validators
from .client import AbstractAsyncClient, AbstractClient, AsyncMiddleware, AsyncMiddlewareHandler, Batch, Middleware
from .client import MiddlewareHandler

__all__ = [
    'AbstractAsyncClient',
    'AbstractClient',
    'AsyncMiddleware',
    'AsyncMiddlewareHandler',
    'Batch',
    'Middleware',
    'MiddlewareHandler',
    'validators',
]



import functools as ft
from typing import Generator

import pytest

from .pytest import PjRpcMocker

PjRpcAiohttpMocker = ft.partial(PjRpcMocker, target='pjrpc.client.backend.aiohttp.Client._request')


@pytest.fixture
def pjrpc_aiohttp_mocker() -> Generator[PjRpcMocker, None, None]:
    """
    Aiohttp client mocking fixture.
    """

    with PjRpcAiohttpMocker() as mocker:
        yield mocker



import json
import typing
from typing import Any, Callable, Generator, Iterable, Mapping, MutableMapping, Optional, TypedDict, Union

import requests.auth
import requests.cookies

import pjrpc
from pjrpc.client import AbstractClient, Middleware
from pjrpc.common import AbstractRequest, AbstractResponse, BatchRequest, BatchResponse, JSONEncoder, JsonRpcError
from pjrpc.common import Request, Response, generators
from pjrpc.common.typedefs import JsonRpcRequestIdT


class RequestArgs(TypedDict, total=False):
    headers: Mapping[str, Union[str, bytes, None]]
    cookies: requests.cookies.RequestsCookieJar
    auth: Union[tuple[str, str], requests.auth.AuthBase]
    timeout: Union[float, tuple[float, float], tuple[float, None]]
    allow_redirects: bool
    proxies: MutableMapping[str, str]
    hooks: Mapping[str, Union[Iterable[Callable[[requests.Response], Any]], Callable[[requests.Response], Any]]]
    verify: Union[bool, str]
    cert: Union[str, tuple[str, str]]


class Client(AbstractClient):
    """
    `Requests <https://2.python-requests.org/>`_ library client backend.

    :param url: url to be used as JSON-RPC endpoint.
    :param session: custom session to be used instead of :py:class:`requests.Session`
    :param id_gen_impl: identifier generator
    :param error_cls: JSON-RPC error base class
    :param json_loader: json loader
    :param json_dumper: json dumper
    :param json_encoder: json encoder
    :param json_decoder: json decoder
    """

    def __init__(
        self,
        url: str,
        *,
        session: Optional[requests.Session] = None,
        raise_for_status: bool = True,
        id_gen_impl: Callable[..., Generator[JsonRpcRequestIdT, None, None]] = generators.sequential,
        error_cls: type[JsonRpcError] = JsonRpcError,
        json_loader: Callable[..., Any] = json.loads,
        json_dumper: Callable[..., str] = json.dumps,
        json_encoder: type[JSONEncoder] = JSONEncoder,
        json_decoder: Optional[json.JSONDecoder] = None,
        middlewares: Iterable[Middleware] = (),
    ):
        super().__init__(
            id_gen_impl=id_gen_impl,
            error_cls=error_cls,
            json_loader=json_loader,
            json_dumper=json_dumper,
            json_encoder=json_encoder,
            json_decoder=json_decoder,
            middlewares=middlewares,
        )
        self._endpoint = url
        self._session = session or requests.Session()
        self._owned_session = session is None
        self._raise_for_status = raise_for_status

    @typing.overload
    def send(self, request: Request, **kwargs: Any) -> Optional[Response]:
        ...

    @typing.overload
    def send(self, request: BatchRequest, **kwargs: Any) -> Optional[BatchResponse]:
        ...

    def send(self, request: AbstractRequest, **kwargs: Any) -> Optional[AbstractResponse]:
        """
        Sends a JSON-RPC request.

        :param request: request instance
        :param kwargs: additional client request argument
        :returns: response instance or None if the request is a notification
        """

        return self._send(request, kwargs)

    def _request(
        self,
        request_text: str,
        is_notification: bool,
        request_kwargs: Mapping[str, Any],
    ) -> Optional[str]:
        """
        Sends a JSON-RPC request.

        :param request_text: request text
        :param is_notification: is the request a notification
        :returns: response text
        """

        request_kwargs = typing.cast(RequestArgs, request_kwargs)

        request_kwargs['headers'] = headers = dict(request_kwargs.get('headers', {}))
        headers['Content-Type'] = self._request_content_type

        resp = self._session.post(self._endpoint, data=request_text, **request_kwargs)
        if self._raise_for_status:
            resp.raise_for_status()
        if is_notification:
            return None

        response_text = resp.text
        content_type = resp.headers.get('Content-Type', '')
        if response_text and content_type.split(';')[0] not in self._response_content_types:
            raise pjrpc.exc.DeserializationError(f"unexpected response content type: {content_type}")

        return response_text

    def close(self) -> None:
        """
        Closes the current http session.
        """

        if self._owned_session:
            self._session.close()

    def __enter__(self) -> 'Client':
        if self._owned_session:
            self._session.__enter__()

        return self

    def __exit__(self, *args: Any) -> None:
        if self._owned_session:
            self._session.__exit__(*args)



import abc
import enum
import json
import pathlib
from typing import Any, Iterable, Mapping

from pjrpc.server import Method


class JSONEncoder(json.JSONEncoder):
    """
    Schema JSON encoder.
    """

    def default(self, o: Any) -> Any:
        if isinstance(o, enum.Enum):
            return o.value

        return super().default(o)


class BaseUI(abc.ABC):
    """
    Base UI.
    """

    @abc.abstractmethod
    def get_static_folder(self) -> pathlib.Path:
        """
        Returns ui statics folder.
        """

    @abc.abstractmethod
    def get_index_page(self, spec_url: str) -> str:
        """
        Returns ui index webpage.

        :param spec_url: specification url.
        """


class Specification(abc.ABC):
    """
    JSON-RPC specification.
    """

    @abc.abstractmethod
    def generate(self, root_endpoint: str, methods: Mapping[str, Iterable[Method]]) -> dict[str, Any]:
        """
        Returns specification schema.

        :param root_endpoint: root endpoint all the methods are served on
        :param methods: methods map the specification is generated for.
                        Each item is a mapping from a endpoint to methods on which the methods will be served
        """



import inspect
from typing import Any, Callable, Optional

from pjrpc.common.typedefs import JsonRpcParamsT


class ValidationError(Exception):
    """
    Method parameters validation error. Raised when parameters validation failed.
    """


ExcludeFunc = Callable[[int, str, Optional[type[Any]], Optional[Any]], bool]
MethodType = Callable[..., Any]


class BaseValidatorFactory:
    """
    Base method parameters validator factory. Uses :py:func:`inspect.signature` for validation.

    :param exclude: a function that decides if the parameters must be excluded
                    from validation (useful for dependency injection)
    """

    def __init__(self, exclude: Optional[ExcludeFunc] = None):
        self._exclude = exclude

    def build(self, method: MethodType) -> 'BaseValidator':
        return BaseValidator(method, self._exclude)


class BaseValidator:
    """
    Base method parameters validator.
    """

    def __init__(self, method: MethodType, exclude: Optional[ExcludeFunc] = None):
        self._method = method
        self._exclude = exclude
        self._signature = self._build_signature(method, exclude)

    def validate_params(self, params: Optional['JsonRpcParamsT']) -> dict[str, Any]:
        """
        Validates params against method signature.

        :param params: parameters to be validated

        :raises: :py:class:`pjrpc.server.validators.ValidationError`
        :returns: bound method parameters
        """

        return self._bind(params).arguments

    def _build_signature(self, method: MethodType, exclude: Optional[ExcludeFunc]) -> inspect.Signature:
        """
        Returns method signature.

        :param method: method to get signature of
        :returns: signature
        """

        signature = inspect.signature(method)

        method_parameters: list[inspect.Parameter] = []
        for idx, param in enumerate(signature.parameters.values()):
            if exclude is None or not exclude(idx, param.name, param.annotation, param.default):
                method_parameters.append(param)

        return signature.replace(parameters=method_parameters)

    def _bind(self, params: Optional['JsonRpcParamsT']) -> inspect.BoundArguments:
        """
        Binds parameters to method.
        :param params: parameters to be bound

        :raises: ValidationError is parameters binding failed
        :returns: bound parameters
        """

        method_args = params if isinstance(params, (list, tuple)) else ()
        method_kwargs = params if isinstance(params, dict) else {}

        try:
            return self._signature.bind(*method_args, **method_kwargs)
        except TypeError as e:
            raise ValidationError(str(e)) from e



from typing import Union

__all__ = [
    'JsonRpcParamsT',
    'JsonRpcRequestIdT',
    'JsonT',
]

JsonT = Union[str, int, float, bool, None, list['JsonT'], tuple['JsonT', ...], dict[str, 'JsonT']]
'''JSON type'''  # for sphinx autodoc

JsonRpcRequestIdT = Union[str, int]
'''JSON-RPC identifier type'''  # for sphinx autodoc

JsonRpcParamsT = Union[list[JsonT], tuple[JsonT, ...], dict[str, JsonT]]
'''JSON-RPC parameters type'''  # for sphinx autodoc



import asyncio
import json
import logging
import typing
import uuid
from typing import Any, Callable, Generator, Iterable, Mapping, Optional, TypedDict

import aio_pika
from aio_pika.abc import AbstractIncomingMessage
from yarl import URL

import pjrpc
from pjrpc.client import AbstractAsyncClient, AsyncMiddleware
from pjrpc.common import AbstractRequest, AbstractResponse, BatchRequest, BatchResponse, JSONEncoder, JsonRpcError
from pjrpc.common import Request, Response, generators
from pjrpc.common.typedefs import JsonRpcRequestIdT

logger = logging.getLogger(__package__)


class ExchangeArgs(TypedDict, total=False):
    type: aio_pika.ExchangeType
    durable: bool
    auto_delete: bool
    internal: bool
    passive: bool
    arguments: aio_pika.abc.Arguments
    timeout: aio_pika.abc.TimeoutType


class QueueArgs(TypedDict, total=False):
    durable: bool
    passive: bool
    auto_delete: bool
    arguments: aio_pika.abc.Arguments
    timeout: aio_pika.abc.TimeoutType


class RequestArgs(TypedDict, total=False):
    headers: aio_pika.abc.HeadersType
    delivery_mode: aio_pika.abc.DeliveryMode
    priority: int
    expiration: aio_pika.abc.DateType
    message_id: str
    timestamp: aio_pika.abc.DateType
    type: str
    user_id: str
    app_id: str


class Client(AbstractAsyncClient):
    """
    `aio_pika <https://docs.aio-pika.com/>`_ based JSON-RPC client.

    :param broker_url: broker connection url
    :param connection: broker connection
    :param exchange_name: exchange to publish requests to. If ``None`` default exchange is used
    :param exchange_args: exchange arguments
    :param routing_key: reply message routing key. If ``None`` queue name is used
    :param result_queue_name: result queue name. If ``None`` random exclusive queue is declared for each request
    :param id_gen_impl: identifier generator
    :param error_cls: JSON-RPC error base class
    :param json_loader: json loader
    :param json_dumper: json dumper
    :param json_encoder: json encoder
    :param json_decoder: json decoder
    """

    def __init__(
        self,
        broker_url: Optional[URL] = None,
        *,
        connection: Optional[aio_pika.abc.AbstractConnection] = None,
        exchange_name: str = "",
        routing_key: str,
        exchange_args: Optional[ExchangeArgs] = None,
        result_queue_name: Optional[str] = None,
        result_queue_args: Optional[QueueArgs] = None,
        id_gen_impl: Callable[..., Generator[JsonRpcRequestIdT, None, None]] = generators.sequential,
        error_cls: type[JsonRpcError] = JsonRpcError,
        json_loader: Callable[..., Any] = json.loads,
        json_dumper: Callable[..., str] = json.dumps,
        json_encoder: type[JSONEncoder] = JSONEncoder,
        json_decoder: Optional[json.JSONDecoder] = None,
        middlewares: Iterable[AsyncMiddleware] = (),
    ):
        assert broker_url or connection, "broker_url or connection must be provided"

        super().__init__(
            id_gen_impl=id_gen_impl,
            error_cls=error_cls,
            json_loader=json_loader,
            json_dumper=json_dumper,
            json_encoder=json_encoder,
            json_decoder=json_decoder,
            middlewares=middlewares,
        )

        if connection is not None:
            self._connection = connection
            self._is_connection_owned = False
        else:
            assert broker_url
            self._connection = aio_pika.connection.Connection(broker_url)
            self._is_connection_owned = True

        self._channel: Optional[aio_pika.abc.AbstractChannel] = None

        self._routing_key = routing_key
        self._exchange_name = exchange_name
        self._exchange_args = exchange_args or {}
        self._exchange: Optional[aio_pika.abc.AbstractExchange] = None

        self._result_queue_name = result_queue_name
        self._result_queue_args: QueueArgs = result_queue_args or {}
        self._result_queue: Optional[aio_pika.abc.AbstractQueue] = None
        self._consumer_tag: Optional[str] = None

        self._futures: dict[str, asyncio.Future[str]] = {}

    async def __aenter__(self) -> 'Client':
        await self.connect()
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self.close()

    async def connect(self) -> None:
        """
        Opens a connection to the broker.
        """

        if self._is_connection_owned:
            await self._connection.connect()

        self._channel = channel = await self._connection.channel()

        if self._exchange_name:
            self._exchange = await channel.declare_exchange(self._exchange_name, **self._exchange_args)

        if self._result_queue_name:
            self._result_queue = await channel.declare_queue(self._result_queue_name, **self._result_queue_args)
            self._consumer_tag = await self._result_queue.consume(self._on_result_message, no_ack=True)

    async def close(self) -> None:
        """
        Closes current broker connection.
        """

        if self._consumer_tag and self._result_queue:
            await self._result_queue.cancel(self._consumer_tag)
            self._consumer_tag = None

        for future in self._futures.values():
            if future.done():
                continue

            future.set_exception(asyncio.CancelledError)

        if self._channel:
            await self._channel.close()
            self._channel = None

        if self._connection and self._is_connection_owned:
            await self._connection.close()

    @typing.overload
    async def send(self, request: Request, **kwargs: Any) -> Optional[Response]:
        ...

    @typing.overload
    async def send(self, request: BatchRequest, **kwargs: Any) -> Optional[BatchResponse]:
        ...

    async def send(self, request: AbstractRequest, **kwargs: Any) -> Optional[AbstractResponse]:
        """
        Sends a JSON-RPC request.

        :param request: request instance
        :param kwargs: additional client request argument
        :returns: response instance or None if the request is a notification
        """

        return await self._send(request, kwargs)

    async def _on_result_message(self, message: AbstractIncomingMessage) -> None:
        if not (correlation_id := message.correlation_id):
            logger.info("message correlation id is missing", message.correlation_id)
            return

        future = self._futures.pop(correlation_id, None)

        if future is None:
            logger.warning("unexpected or outdated message received: %r", message)
            return

        if message.content_type not in self._response_content_types:
            future.set_exception(
                pjrpc.exc.DeserializationError(f"unexpected response content type: {message.content_type}"),
            )
        else:
            future.set_result(message.body.decode(message.content_encoding or 'utf8'))

    async def _request(
        self,
        request_text: str,
        is_notification: bool,
        request_kwargs: Mapping[str, Any],
    ) -> Optional[str]:
        assert self._channel, "server is not started"

        request_kwargs = typing.cast(RequestArgs, request_kwargs)

        if is_notification:
            message = aio_pika.message.Message(
                body=request_text.encode(),
                content_encoding='utf8',
                content_type=self._request_content_type,
                **request_kwargs,
            )
            exchange = self._exchange or self._channel.default_exchange
            await exchange.publish(message, routing_key=self._routing_key)
            return None

        request_id = str(uuid.uuid4())

        if not self._result_queue:
            result_queue = await self._channel.declare_queue(request_id, exclusive=True, **self._result_queue_args)
            await result_queue.consume(self._on_result_message, no_ack=True)
        else:
            result_queue = self._result_queue

        message = aio_pika.message.Message(
            body=request_text.encode(),
            correlation_id=request_id,
            reply_to=result_queue.name,
            content_encoding='utf8',
            content_type=self._request_content_type,
            **request_kwargs,
        )

        future: asyncio.Future[str] = asyncio.Future()
        self._futures[request_id] = future

        try:
            exchange = self._exchange or self._channel.default_exchange
            await exchange.publish(message, routing_key=self._routing_key)
            return await future
        finally:
            self._futures.pop(request_id, None)



import json
from typing import Any

from . import exceptions
from .request import BatchRequest, Request
from .response import BatchResponse, Response


class JSONEncoder(json.JSONEncoder):
    """
    Library default JSON encoder. Encodes request, response and error objects to be json serializable.
    All custom encoders should be inherited from it.
    """

    def default(self, o: Any) -> Any:
        if isinstance(
            o, (
                Response, Request,
                BatchResponse, BatchRequest,
                exceptions.JsonRpcError,
            ),
        ):
            return o.to_json()

        return super().default(o)
